/*
 * arch/mips/tango2/delay.c
 *
 * Copyright (C) 2003-2004 Sigma Designs, Inc
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/tango2/hardware.h>
#include <asm/tango2/tango2api.h>
#include <asm/io.h>
#include <linux/irq.h>
#include <linux/sched.h>

static inline unsigned long tango2_getxtal(void)
{
	return(gbus_read_uint32(pGBus, REG_BASE_system_block + SYS_xtal_in_cnt));
}

/* This is the replacement of Linux's udelay. */
void tango2_udelay(unsigned usec)
{
	/* SYS_xtal_in_cnt is a counter running off 27MHz, so 1 usec
           is roughly equivalent to 27 increase of count */
	unsigned long start = tango2_getxtal();
	unsigned long end = start + (usec * 27);

	if (end <= start) 
		/* Handle overflow condition */
		while (tango2_getxtal() > start);
	while (tango2_getxtal() < end);
}

void tango2_syncwith_xtal(unsigned long *mark, unsigned usec)
{
	unsigned long end = *mark + (usec * 27);
	if (end > *mark)
		/* Handle overflow condition */
		while (tango2_getxtal() > *mark);
	while (tango2_getxtal() < end);
	*mark = end;
}

EXPORT_SYMBOL(tango2_udelay);

