/*
	Refer to bug #3644.
	
	TLB-based implementation works on the ranges:
	[0x00000000..0x20000000[ access thru kseg1
	[0x20000000..0x2xxxxxxx (size of dram1, a parameter of ioremap())[ access thru tlb. Outside: unpredictable/oops.

	Remap-based implementation does:
	00xy b27b26..b0 to 101y(b27|x)b26..b0.
       and works everywhere excepted ranges: [0x18000000..0x20000000[ and [0x28000000..0xffffffff]

	Test with
	{
		volatile int q=gbus_read_uint32(pGBus,0x1020212c); // correct
		q=gbus_read_uint32(pGBus,0x1f20212c);              // incorrect
		q=gbus_read_uint32(pGBus,0x2020212c);              // correct
		q=gbus_read_uint32(pGBus,0x2720212c);              // correct
		q=gbus_read_uint32(pGBus,0x2820212c);              // incorrect
		q=gbus_read_uint32(pGBus,0x2f20212c);              // incorrect
	}
*/
	
#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/mipsregs.h>

.set noreorder
.set mips32r2

#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1

LEAF(gbus_read_dram_uint8)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	lbu v0, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	lbu v0, 0(t0)
END(gbus_read_dram_uint8)

LEAF(gbus_read_dram_uint16)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	lhu v0, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	lhu v0, 0(t0)
END(gbus_read_dram_uint16)

LEAF(gbus_read_dram_uint32)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	lw v0, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	lw v0, 0(t0)
END(gbus_read_dram_uint32)

LEAF(gbus_write_dram_uint8)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	sb a2, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	sb a2, 0(t0)
END(gbus_write_dram_uint8)

LEAF(gbus_write_dram_uint16)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	sh a2, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	sh a2, 0(t0)
END(gbus_write_dram_uint16)

LEAF(gbus_write_dram_uint32)
	ext t0, a1, 29, 1 // 0x20000000 is bit29
	bnez t0, 1f
	lui t1, 0xa000
	or t0, t1, a1
	jr ra
	sw a2, 0(t0)
1:
	lui t1, 0x2000
	sub t0, a1, t1
	lw t1, em86xx_tlb_dram1_map_base
	add t0, t1
	jr ra
	sw a2, 0(t0)
END(gbus_write_dram_uint32)

#else
		
LEAF(gbus_read_dram_uint8)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	lbu v0, 0(t0)
END(gbus_read_dram_uint8)
	
LEAF(gbus_read_dram_uint16)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	lhu v0, 0(t0)
END(gbus_read_dram_uint16)
	
LEAF(gbus_read_dram_uint32)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	lw v0, 0(t0)
END(gbus_read_dram_uint32)
	
LEAF(gbus_write_dram_uint8)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	sb a2, 0(t0)
END(gbus_write_dram_uint8)
	
LEAF(gbus_write_dram_uint16)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	sh a2, 0(t0)
END(gbus_write_dram_uint16)
	
LEAF(gbus_write_dram_uint32)
	rotr t0, a1, 29
	ins t0, t0, 30, 1
	or t0, 5
	rotr t0, 3
	jr ra
	sw a2, 0(t0)
END(gbus_write_dram_uint32)

#endif
