/*****************************************
 Copyright © 2001-2003  
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file   pci.c
  @brief  

  PCI support for Tango2.

  @author YH Lin
  @date   2005-01-20
*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/module.h>
                                                                                
#include <asm/pci_channel.h>
#include <asm/debug.h>


#include <asm/tango2/hardware.h>

#define PCIEM86XX_ARBITER_LEVEL         0x00
#define PCIEM86XX_ARBITER_GRANTTIMEOUT  0x10101010

#define PCIEM86XX_IDSEL_MAX	0x05
#define PCIEM86XX_IDSEL_BITS	0x03

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
extern int tango2_pci_host_enabled(void);
extern int tango2_pcidev_enabled(const int pci_idsel);
#endif

static int tango2_pci_read_config_byte(struct pci_dev *dev, int where, u8 *value);
static int tango2_pci_read_config_word(struct pci_dev *dev, int where, u16 *value);
static int tango2_pci_read_config_dword(struct pci_dev *dev, int where, u32 *value);
static int tango2_pci_write_config_byte(struct pci_dev *dev, int where, u8 value);
static int tango2_pci_write_config_word(struct pci_dev *dev, int where, u16 value);
static int tango2_pci_write_config_dword(struct pci_dev *dev, int where, u32 value);
                                                                                
struct pci_ops tango2_ext_pci_ops = {
    tango2_pci_read_config_byte,
    tango2_pci_read_config_word,
    tango2_pci_read_config_dword,
    tango2_pci_write_config_byte,
    tango2_pci_write_config_word,
    tango2_pci_write_config_dword,
};

static struct resource extpci_io_resource = {
	"PCI IO space",
	MEMORY_BASE_PCI_IO,
	MEMORY_BASE_PCI_IO + 0x08000000 - 1,
	IORESOURCE_IO
};

static struct resource extpci_mem_resource = {
	"PCI MEM space",
	MEMORY_BASE_PCI_MEMORY + 0x02000000,
	MEMORY_BASE_PCI_MEMORY + 0x20000000 - 1,
	IORESOURCE_MEM
};

struct pci_channel mips_pci_channels[] = {
	{ &tango2_ext_pci_ops, &extpci_io_resource, &extpci_mem_resource },
	{ NULL, NULL, NULL }
};

static int g_last_idsel = -1;
static int g_idsel_exist[PCIEM86XX_IDSEL_MAX] = { 0, 0, 0, 0, 0 };
static int g_idsel_func_exist[PCIEM86XX_IDSEL_MAX * 8];

extern unsigned long em8xxx_kmem_start;
extern unsigned long em8xxx_kmem_size;

unsigned long g_pcimem_busaddr = PCIBIOS_MIN_MEM_EM86XX - MEMORY_BASE_PCI_MEMORY;
unsigned long g_pcimem_phyaddr = 0;
unsigned long g_pcimem_phyaddr_end = 0;

unsigned long tango2_pci_config_base = 0;
static unsigned long tango2_pci_io_base = 0;
//static unsigned long tango2_pci_mem_base = 0;

#define PCIEM86XX_CONFIG_START                                              \
    unsigned long flags;                                                    \
    int idsel = PCI_SLOT(dev->devfn), func = PCI_FUNC(dev->devfn);          \
                                                                            \
    if (idsel >= PCIEM86XX_IDSEL_MAX || !g_idsel_exist[idsel] || !g_idsel_func_exist[idsel * 8 + func])              \
        return PCIBIOS_DEVICE_NOT_FOUND;                                    \
    local_irq_save(flags);
                                                                                
#define PCIEM86XX_CONFIG_END                                                \
    local_irq_restore(flags);                                               \
    return PCIBIOS_SUCCESSFUL;
                                                                                
#define PCIEM86XX_CONFIG_READ(x, type, size)                                \
    static int tango2_pci_read_config_##type(struct pci_dev *dev, int where, u##size *value) \
    {                                                                       \
        PCIEM86XX_CONFIG_START                                              \
        *value = __raw_read##x(tango2_pci_config_base + ((func << 8) | where | (idsel << 11) | (dev->bus->number << 16)));\
        {                                                   \
            unsigned long data = (gbus_read_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3) >> 1) & 0x3;   \
            if (data) { /* bus fault */                                     \
if (0) printk("PCI Bus fault detected 0x%x\n", data); \
		g_idsel_func_exist[idsel * 8 + func] = 0; \
                gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 1); \
                gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 0); \
                return PCIBIOS_DEVICE_NOT_FOUND;                            \
            }                                                               \
        }                                                                   \
        PCIEM86XX_CONFIG_END                                                \
    }
                                                                                
#define PCIEM86XX_CONFIG_WRITE(x, type, size)                               \
    static int tango2_pci_write_config_##type(struct pci_dev *dev, int where, u##size value) \
    {                                                                       \
        PCIEM86XX_CONFIG_START                                              \
        __raw_write##x(value, tango2_pci_config_base + ((func << 8) | where | (idsel << 11) | (dev->bus->number << 16)));    \
        PCIEM86XX_CONFIG_END \
    }
                                                                                
PCIEM86XX_CONFIG_READ(b, byte, 8)
PCIEM86XX_CONFIG_READ(w, word, 16)
PCIEM86XX_CONFIG_READ(l, dword, 32)
                                                                                
PCIEM86XX_CONFIG_WRITE(b, byte, 8)
PCIEM86XX_CONFIG_WRITE(w, word, 16)
PCIEM86XX_CONFIG_WRITE(l, dword, 32)

static void tango2_pci_busfault_intr(int irq, void *devinfo, struct pt_regs *regs);

// return 0 if device not found
// return vendorid:deviceid if device found
static unsigned int tango2_pci_select(int idsel)
{
    unsigned int data, pciid;
                                                                                
    g_last_idsel = idsel;
    gbus_write_uint16(pGBus, REG_BASE_host_interface + PCI_host_reg2, idsel << 8);
    pciid = __raw_readl(tango2_pci_config_base);
    data = (gbus_read_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3) >> 1) & 0x3;
                                                                                
    if (data == 0) {
        return pciid;
    } else {
        // clears PCI bus fault flags
        gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 1);
        gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 0);
        return 0;
    }
}

static int __init tango2_pci_init(void)
{
	int i, idsel; 
	unsigned pciid = 0;

    // Set to PCI host
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_chip_is_host, 1);

    // HOST_REG1 :
    //   [31:16] : # of PCI retry cycle = 0xffff (default = 0xff)
    //   [8] : host Super Request = 0
    //   [3:0] arbitration level = 0x00 (Level 1)
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_host_reg1,
		0xffff0000 | PCIEM86XX_ARBITER_LEVEL);

    // PCI_CTRL1 :
    //   [17] : enable "Memory Read Multiple" and "Memory Read Line"
    //   [16] [7:0] : enable "prefetch" for PCI slave regions 2..7
    //   [17] : Always enabled
    //   [18] : additional bit for Tango. Long PCI memory read burst
//	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_pcictrl_reg1, 0x000700fc);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_pcictrl_reg1, 0x00030000);

    // PCI_CTRL2 :
    //   [18] : fast back-to-back capable = 0 (default)
    //   [17] : read FIFO level = 1 (8 level deep, default)
    //   [16] : discard timer enable = 1 (default)
    //   [15:8] : subs latency = 0x06 (default = 0x08)
    //   [7:0] : initial latency = 0x0d (default = 0x0b)
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_pcictrl_reg2, 0x0003060d);

    // PCI_CTRL3 :
    //   [16] : slave abort clear = 0
    //   [10:8] : abort interrupt enable = 0 (default)
    //   [2:0] : abort status = 0
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_pcictrl_reg3, 0);

    // clear PCI bus fault
    if (((gbus_read_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3) >> 1) & 0x3) != 0) {
	gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 1);
	gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 0);
    }

    // Setting pci_configuration_vld 
    gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 2, 1);

    // grant timeout
    gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_host_reg3, PCIEM86XX_ARBITER_GRANTTIMEOUT);
    gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_host_reg5, PCIEM86XX_ARBITER_GRANTTIMEOUT & 0x000000ff);
                                                                                
    // initialize arbiter
    gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_host_reg4, 0);

    // detect each agent
    memset(&g_idsel_func_exist[0], 0, sizeof(int) * PCIEM86XX_IDSEL_MAX * 8);

    /* Briefly scanning the bus for devices */
    for (idsel = 1; idsel < PCIEM86XX_IDSEL_MAX; ++idsel) {
        for (i = 1; i <= 3; ++i) {
            if ((pciid = tango2_pci_select(idsel)) != 0)
                break;
        }
#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
    	if (tango2_pcidev_enabled(idsel) == 0)
      	    g_idsel_exist[idsel] = pciid = 0;
	else
            g_idsel_exist[idsel] = (pciid == 0) ? 0 : 1;
#else
        g_idsel_exist[idsel] = (pciid == 0) ? 0 : 1;
#endif
	if (pciid != 0) {
	    printk("PCI: found device IDSEL = %d\n", idsel);
	    for (i = idsel * 8; i < (idsel + 1) * 8; i++)
		g_idsel_func_exist[i] = 1;
	}
    }

    // Initialize IDSEL 0 - EM86XX as a PCI slave
    for (i = 1; i <= 3; ++i) {
        if ((pciid = tango2_pci_select(0)) != 0)
            break;
        printk("PCI: Probing EM86XX at IDSEL 0, retry = %d\n", i);
    }

    if (pciid == 0) {
        printk("PCI: Can't initialize EM86XX as a PCI slave\n");
	goto pci_init_end;
    } else {
        /* Start configuring memory spaces .. */
        unsigned long memsize, regsize, membase;

	g_pcimem_phyaddr = PHYSADDR(em8xxx_kmem_start);

        // set PCI memory size to maximum, so the PCI memory will cover the
        // whole memory if the total DRAM size is smaller than 96MB.
        // maximum = 128MB => 16MB per region => DMA memory = 16 * 6 = 96MB
        // But the current code doesn't care 2nd DRAM controller.
        gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_REG3, 0x7);
                                                                                
        memsize = gbus_read_uint32(pGBus, REG_BASE_host_interface + PCI_REG3) & 0x07;
        memsize = 1 << memsize;
        regsize = (memsize << 20) >> 3; // memory size / 8

        // pci command
        __raw_writew(PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER, tango2_pci_config_base + PCI_COMMAND);

        // base address 0
        __raw_writel(g_pcimem_busaddr, tango2_pci_config_base + PCI_BASE_ADDRESS_0);
                                                                                
        // PCI slave access
        // region 0 (R) : Configuration area
        // region 1 : RISC memory
        gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_REGION_0_BASE + (1 * 4), MEM_BASE_mpeg_engine_0);
                                                                                
        // region 2 - : mapped to DRAM starting from DRAM_BASE
        g_pcimem_busaddr += regsize * 2;
        printk("PCI: Configured SMP863x as PCI slave with %dMB PCI memory\n", memsize);
        membase = PHYSADDR(em8xxx_kmem_start + 0x00000000);
        g_pcimem_phyaddr = membase;
        g_pcimem_phyaddr_end = g_pcimem_phyaddr + em8xxx_kmem_size;
        for (i = 2; i < 8; ++i) {
            // PCI region base registers should contain low address
            gbus_write_uint32(pGBus, REG_BASE_host_interface + PCI_REGION_0_BASE + (i * 4), PHYSADDR(membase));
            membase += regsize;
        }
        if (g_pcimem_phyaddr_end > membase)
            g_pcimem_phyaddr_end = membase;
        printk("PCI: Region size is %dKB\n", regsize >> 10);
        printk("PCI: Map DMA memory 0x%08x-0x%08x for PCI at 0x%08x\n",
            g_pcimem_phyaddr, g_pcimem_phyaddr_end, g_pcimem_busaddr);
    }

pci_init_end:

    return(0);
}

static void tango2_pci_busfault_intr(int irq, void *devinfo, struct pt_regs *regs)
{
    static char *s_fault_reason[] = {
        "OK", "Master Abort", "Retry timer expired", "Unknown" };
                                                                                
    static int s_faultcount = 0;
    unsigned int data;
                                                                                
    data = (gbus_read_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3) >> 1) & 3;
                                                                                
    if (data == 1) {
        printk("GPF: pid(%d, <%s>), EPC=0x%08lx\n", current->pid, current->comm, regs->cp0_epc);
    } else
        printk("PCI: PCI bus fault (%d): %s", ++s_faultcount, s_fault_reason[data]);

    // clears PCI bus fault flags
    gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 1);
    gbus_write_uint8(pGBus, REG_BASE_host_interface + PCI_host_reg2 + 3, 0);
}

void __init tango2_pcibios_init(void)
{
#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
	if (tango2_pci_host_enabled() == 0)
		return;
#endif

	tango2_pci_config_base = (unsigned long)ioremap(MEMORY_BASE_PCI_CONFIG, PCIEM86XX_IDSEL_MAX * 0x800);
	tango2_pci_io_base = (unsigned long)ioremap(MEMORY_BASE_PCI_IO, 0x08000000);
//	tango2_pci_mem_base = (unsigned long)ioremap(MEMORY_BASE_PCI_MEMORY, 0x10000000); /* should be 0x20000000 */

	printk("Remapping PCI config space 0x%08lx to 0x%08lx, size 0x%x\n", 
		MEMORY_BASE_PCI_CONFIG, tango2_pci_config_base, PCIEM86XX_IDSEL_MAX * 0x800);
	printk("Remapping PCI I/O space 0x%08lx to 0x%08lx, size 0x%x\n", 
		MEMORY_BASE_PCI_IO, tango2_pci_io_base, 0x08000000);
//	printk("Remapping PCI memory space 0x%08lx to 0x%08lx, size 0x%x\n", 
//		MEMORY_BASE_PCI_MEMORY, tango2_pci_mem_base, 0x10000000);

	/* The base address for MIPS I/O port, which will be used by
	   I/O functions such like inb(), outb(), ...  */
	set_io_port_base(tango2_pci_io_base);

	tango2_pci_init();

        request_irq(LOG2_CPU_PCI_FAULT_INT + IRQ_CONTROLLER_IRQ_BASE, tango2_pci_busfault_intr, SA_INTERRUPT, "TANGO2 PCI HOST", NULL);
}

unsigned long tango2_virt_to_bus(unsigned long virt_addr)
{
    if ((PHYSADDR(virt_addr) < g_pcimem_phyaddr) || (PHYSADDR(virt_addr) >= g_pcimem_phyaddr_end))
	printk("virt2bus: Not a dma-able address: 0x%08x\n", virt_addr); 
    return((unsigned long)(PHYSADDR(virt_addr)-g_pcimem_phyaddr+g_pcimem_busaddr));
}

unsigned long tango2_bus_to_virt(unsigned long bus_addr)
{
    if ((bus_addr < g_pcimem_busaddr) || (bus_addr >= (g_pcimem_busaddr + (g_pcimem_phyaddr_end - g_pcimem_phyaddr)))) 
	printk("bus2virt: Not a valid bus address: 0x%08x\n", bus_addr);
    return((unsigned long)(bus_addr-g_pcimem_busaddr+g_pcimem_phyaddr));
}

EXPORT_SYMBOL(tango2_virt_to_bus);
EXPORT_SYMBOL(tango2_bus_to_virt);

