/*
 * Copyright 2001 MontaVista Software Inc.
 * Author: Jun Sun, jsun@mvista.com or jsun@junsun.net
 *
 * arch/mips/tango2/prom.c
 *     prom setup file for tango2
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 ***********************************************************************
 */

#include <linux/init.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/bootmem.h>

#include <asm/bootinfo.h>
#include <asm/addrspace.h>
#include <asm/wbflush.h>

#include <asm/tango2/tango2.h>

#include <asm/tango2/memcfg.h>
#include <asm/tango2/sigblock.h>
#include <asm/tango2/zboot.h>

#ifdef CONFIG_TANGO2_XENV
#include "xenv.h"
#include "xenvkeys.h"
#endif

char arcs_cmdline[CL_SIZE];
unsigned long em8xxx_cpu_frequency = 0L;
unsigned long em8xxx_sys_frequency = 0L;
unsigned long em8xxx_kmem_start = 0L;
unsigned long em8xxx_kmem_size = 0L;

#if defined(CONFIG_TANGO2_SIG_BLOCK)
struct signature_block *sigptr = NULL;
static struct signature_block sigblock;
#elif defined(CONFIG_TANGO2_XENV)

#define REMAPPED_REG	CPU_remap4
#define REMAPPED_BASE	KSEG1ADDR(CPU_remap4_address) 

#define XENV_MAX_FLASH_PARTITIONS   16

unsigned int xenv_scard_5v_pin = 0, xenv_scard_off_pin = 0, xenv_scard_cmd_pin = 0;
unsigned int xenv_cs_flash_size[4] = { 0, 0, 0, 0 };
unsigned int xenv_cs_flash_parts[4] = { 0, 0, 0, 0 };
unsigned int xenv_flash_parts_size[4][XENV_MAX_FLASH_PARTITIONS];
unsigned int xenv_flash_parts_offset[4][XENV_MAX_FLASH_PARTITIONS];
unsigned int xenv_uart_console_port = 0;
unsigned int xenv_uart_used_ports = 1;

extern void tango2_init_uart_ports(void);

static unsigned int xenv_enabled_devices = 0;
static unsigned int xenv_isaide_timing_slot = 0;
static unsigned int xenv_isaide_irq = 0;
static unsigned int xenv_mac_hi = 0, xenv_mac_lo = 0;
static unsigned int xenv_pcidev_irq_route[4] = { 0, 0, 0, 0 };

static unsigned long xenv_addr = 0, xenv_gbus_addr = 0;

#define ISAIDE_SHIFT		0
#define BMIDE_SHIFT		1
#define PCIHOST_SHIFT		2
#define ETHERNET_SHIFT		3
#define IR_SHIFT		4
#define FIP_SHIFT		5	
#define I2CM_SHIFT		6
#define I2CS_SHIFT		7
#define SDIO_SHIFT		8
#define USB_SHIFT		9
#define PCI1_SHIFT		10
#define PCI2_SHIFT		11
#define PCI3_SHIFT		12
#define PCI4_SHIFT		13
#define PCI5_SHIFT		14
#define PCI6_SHIFT		15
#define SATA_SHIFT		16
#define SCARD_SHIFT		17
#define GNET_SHIFT		18
/*				19-32: undefined */
#else
/* !CONFIG_TANGO2_SIG_BLOCK && !CONFIG_TANGO2_XENV */
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
static int tango2_device_info(void);
unsigned long em8xxx_baudrate = CONFIG_TANGO2_BASE_BAUD;
#else
/* !CONFIG_TANGO2_SIG_BLOCK && !CONFIG_TANGO2_XENV */
#endif

#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
unsigned long em86xx_tlb_dram1_map_base = 0L;
unsigned long em86xx_tlb_dram1_map_size = 0L;
#endif

const char *get_system_type(void)
{
	return "SigmaDesigns Tango2";
}

unsigned long tango2_get_sysclock(void)
{
#if (EM86XX_REVISION > 3)
	unsigned long sys_clkgen_pll, sysclk_mux, sysclk_premux, n, m, freq, div, k, mux;

	k = m = sys_clkgen_pll = 0;
 	sysclk_mux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_mux);
 	sysclk_premux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_premux);
	switch(sysclk_premux & 0x3) {
		case 0:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen0_pll);
			m = (sys_clkgen_pll >> 16) & 0x1f;
			k = (sys_clkgen_pll >> 14) & 0x3;
			break;
		case 1:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen1_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
		case 2:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen2_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
		case 3:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen3_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
	}
        n = sys_clkgen_pll & 0x000003ff;

	/* Not using XTAL_IN, cannot calculate */
	if ((sys_clkgen_pll & 0x07000000) != 0x01000000)
		return(0);

	/* Calculate the divider */
	mux = (sysclk_mux >> 8) & 0xf;
	if (mux == 0) /* Get system clock frequency */
		div = 2;
	else if ((mux == 1) || ((mux >= 8) && (mux < 0xc)))
		div = 4;
	else if ((mux >= 2) && (mux < 8))
		div = 3;
	else
		return(0); /* Wrong divider setting */

	if (sysclk_mux & 0x1) 	/* PLL is used */
		freq = ((TANGO2_BASE_FREQUENCY / (m + 2)) * (n + 2)) / (div * (1 << k));
	else
		freq = TANGO2_BASE_FREQUENCY / div;
#else
	unsigned long sys_clkgen_pll, sysclk_mux, n, m, freq, div;
 	sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen_pll);
 	sysclk_mux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_mux);
        n = sys_clkgen_pll & 0x000003ff;
        m = (sys_clkgen_pll & 0x003f0000) >> 16;

	/* Calculate the divider */
	if ((sysclk_mux & 0x300) == 0x000) /* Get system clock frequency */
		div = 2;
	else if ((sysclk_mux & 0x300) == 0x100)
		div = 4;
	else
		div = 3;

	if (sysclk_mux & 0x1) 	/* PLL is used */
		freq = ((TANGO2_BASE_FREQUENCY / (m + 2)) * (n + 2)) / div;
	else
		freq = TANGO2_BASE_FREQUENCY / div;
#endif
	return(freq);
}

unsigned long tango2_get_cpuclock(void)
{
#if (EM86XX_REVISION > 3)
	unsigned long sys_clkgen_pll, sysclk_mux, sysclk_premux, n, m, freq, div, k, mux;

	k = m = sys_clkgen_pll = 0;
 	sysclk_mux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_mux);
 	sysclk_premux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_premux);
	switch(sysclk_premux & 0x3) {
		case 0:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen0_pll);
			m = (sys_clkgen_pll >> 16) & 0x1f;
			k = (sys_clkgen_pll >> 14) & 0x3;
			break;
		case 1:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen1_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
		case 2:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen2_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
		case 3:
			sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen3_pll);
			m = (sys_clkgen_pll >> 16) & 0x7f;
			break;
	}
        n = sys_clkgen_pll & 0x000003ff;

	/* Not using XTAL_IN, cannot calculate */
	if ((sys_clkgen_pll & 0x07000000) != 0x01000000)
		return(0);

	/* Calculate the divider */
	mux = (sysclk_mux >> 8) & 0xf;
	if ((mux == 3) || (mux == 4) || (mux == 6)) /* Get CPU frequency */
		div = 3;
	else if ((mux == 8) || (mux == 0xa))
		div = 4;
	else if ((mux == 0) || (mux == 1) || (mux == 2) || (mux == 5) || (mux == 7) || 
			(mux == 9) || (mux == 0xb))
		div = 2;
	else
		return(0); /* Wrong divider setting */

	if (sysclk_mux & 0x1) 	/* PLL is used */
		freq = ((TANGO2_BASE_FREQUENCY / (m + 2)) * (n + 2)) / (div * (1 << k));
	else
		freq = TANGO2_BASE_FREQUENCY / div;
#else
	unsigned long sys_clkgen_pll, sysclk_mux, n, m, freq, div;
 	sys_clkgen_pll = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_clkgen_pll);
 	sysclk_mux = *(volatile unsigned long *)KSEG1ADDR(REG_BASE_system_block + SYS_sysclk_mux);
        n = sys_clkgen_pll & 0x000003ff;
        m = (sys_clkgen_pll & 0x003f0000) >> 16;

	if ((sysclk_mux & 0x300) == 0x300) /* Get CPU frequency */
		div = 3;
	else
		div = 2;

	/* Calculate the divider */
	if (sysclk_mux & 0x1) 	/* PLL is used */
		freq = ((TANGO2_BASE_FREQUENCY / (m + 2)) * (n + 2)) / div;
	else
		freq = TANGO2_BASE_FREQUENCY / div;
#endif
	return(freq);
}

#if 0
// rather do this from gdb with jtag [hwb_rw gdb macro]

/// 
/**
   <long-description>

   @param start 
   @param size  power of two bw 8 and 4096
   @param mode  1:W 2:R 4:I
*/
void setup_watch(unsigned long start,int size,int mode)
{
	start=start&(~7);

	emprintk("watching mode %d [%p,%p]\n",mode,start,start+size);
	write_c0_watchlo0(start|mode);
	write_c0_watchhi0((1<<30)|((size-1)&(~7)));
}
#endif

void kernel_entry(void);

int __init prom_init(int argc, char **argv, char **envp, int *prom_vec)
{
	extern char _ftext; /* For code start address (not entry address)  */
	unsigned long kend;
	unsigned long offset;
	memcfg_t *m=(memcfg_t *)KSEG1ADDR(MEM_BASE_dram_controller_0+FM_MEMCFG);
	int i;

#ifdef CONFIG_CMDLINE_BOOL
	strcpy (arcs_cmdline, CONFIG_CMDLINE);
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK)
	/* Save the signature block locally */
	memcpy(&sigblock, (struct signature_block *)KSEG1ADDR(0x1fc00000), sizeof(struct signature_block));
	sigptr = &sigblock;

	/* Get the baudrate information from signature block */
	if ((em8xxx_baudrate = zboot_default_baudrate(sigptr)) == 0) 
		em8xxx_baudrate = CONFIG_TANGO2_BASE_BAUD; /* Use default */
#elif defined(CONFIG_TANGO2_XENV)
	if ((xenv_gbus_addr = xenv_addr = gbus_read_uint32(pGBus, REG_BASE_cpu_block + LR_XENV_LOCATION)) != 0) {
		printk("XENV_ADDR=0x%08x\n", xenv_addr);
		/* Got the xenv address in gbus form, now convert it in remap from so we can access it */
		gbus_write_uint32(pGBus, REG_BASE_cpu_block + REMAPPED_REG, xenv_addr & 0xfc000000);
		iob();
		xenv_addr = REMAPPED_BASE + (xenv_addr & 0x03ffffff);

		if (xenv_isvalid((void *)xenv_addr, MAX_XENV_SIZE) < 0)
			xenv_addr = 0;
	}

	if (xenv_addr == 0) 
		printk("WARNING: Cannot found valid XENV block.\n");
	else {
		extern int tango2_pcidev_enabled(const int pci_idsel);
		unsigned long cs_config = (gbus_read_uint32(pGBus, REG_BASE_host_interface + PB_CS_config) >> 12) & 0xf;
		unsigned int val;
		RMuint32 size;
		char buf[32], xenv_cmdline[CL_SIZE];
		int j;

		memset(&xenv_flash_parts_size[0][0], 0, sizeof(xenv_flash_parts_size));
		memset(&xenv_flash_parts_offset[0][0], 0, sizeof(xenv_flash_parts_offset));

		/* Getting Linux command line from XENV */
		size = CL_SIZE;
		if (xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_LINUX_CMD, xenv_cmdline, &size) == RM_OK) {
			xenv_cmdline[size] = '\0';
			if (xenv_cmdline[0] == '\"') {
				if (xenv_cmdline[size -1] == '\"')
					xenv_cmdline[size - 1] = '\0';
				strcpy(arcs_cmdline, &xenv_cmdline[1]);
			} else
				strcpy(arcs_cmdline, &xenv_cmdline[0]);
		}

		size = 4;
		if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_DEF_BAUDRATE, &val, &size) == RM_OK) && (size == 4))
			em8xxx_baudrate = val;
		size = 4;
		if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_ENABLED_DEVICES, &val, &size) == RM_OK) && (size == 4)) 
			xenv_enabled_devices = val;
		size = 4;
		if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_ISAIDE_IRQ_ROUTE, &val, &size) == RM_OK) && (size == 4)) 
			xenv_isaide_irq = (int)(val & 0x3);
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_ISAIDE_TIMING_SLOT, &val, &size) == RM_OK) && (size == 4))
			xenv_isaide_timing_slot = (int)(val & 0x7);
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_SCARD_OFF, &val, &size) == RM_OK) && (size == 4))
			xenv_scard_off_pin = val;
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_SCARD_5V, &val, &size) == RM_OK) && (size == 4))
			xenv_scard_5v_pin = val;
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_SCARD_CMD, &val, &size) == RM_OK) && (size == 4))
			xenv_scard_cmd_pin = val;
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_CONSOLE_UART_PORT, &val, &size) == RM_OK) && (size == 4))
			xenv_uart_console_port = val;
		size = 4;
	        if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, XENV_KEY_UART_USED_PORTS, &val, &size) == RM_OK) && (size == 4))
			xenv_uart_used_ports = val;

		/* Boot program (e.g. zboot) should set up this already. */
		xenv_mac_hi = gbus_read_uint32(pGBus, REG_BASE_cpu_block + LR_ETH_MAC_HI) & 0xffff;
		xenv_mac_lo = gbus_read_uint32(pGBus, REG_BASE_cpu_block + LR_ETH_MAC_LO);

		// Getting PCI IRQ routing information
		for (i = 1; i <= 4; i++) {
			if (!tango2_pcidev_enabled(i))
				continue;
			sprintf(buf, XENV_KEYS_PCI_IRQ_ROUTE, i);
	                size = 4;
	                if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, buf, &val, &size) == RM_OK) && (size == 4)) 
				xenv_pcidev_irq_route[i - 1] = val;
		}
		/* Save all cs0-cs3 flash keys needed from XENV */
		for (i = 0; i < 4; i++) {
			if ((cs_config & (1 << i)) != 0) /* Check to see if see to IDE mode */
				continue;
			sprintf(buf, XENV_KEYS_CS_SIZE, i);
			size = 4;
			if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, buf, &val, &size) == RM_OK) && (size == 4)) 
				xenv_cs_flash_size[i] = val;
			if (xenv_cs_flash_size[i] == 0)
				continue;
			sprintf(buf, XENV_KEYS_CS_PARTS, i);
			size = 4;
			if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, buf, &val, &size) == RM_OK) && (size == 4)) {
				xenv_cs_flash_parts[i] = val;
				for (j = 0; j < xenv_cs_flash_parts[i]; j++) {
					sprintf(buf, XENV_KEYS_CS_PART_SIZE, i, j + 1);
					size = 4;
					if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, buf, &val, &size) == RM_OK) && (size == 4)) 
						xenv_flash_parts_size[i][j] = val;
					sprintf(buf, XENV_KEYS_CS_PART_OFFSET, i, j + 1);
					size = 4;
					if ((xenv_get((void *)xenv_addr, MAX_XENV_SIZE, buf, &val, &size) == RM_OK) && (size == 4)) 
						xenv_flash_parts_offset[i][j] = val;
				}
			}
		}

		/* Call out to initialize serial ports data */
		tango2_init_uart_ports();
	}
#endif

	/* Pass argument in ordinary way, override the command line */
	if ((argc > 1) && (argv != NULL)) {
		arcs_cmdline[0] = '\0';
		for (i = 1; i < argc; i++) {
			strcat(arcs_cmdline, argv[i]);
			strcat(arcs_cmdline, " ");
		}
	}

	em8xxx_kmem_start = ((unsigned long)(&_ftext) & 0xfffff000); /* Aligned by 4KB */
	/* Aligned by MB boundary */
	em8xxx_kmem_size = ((em8xxx_kmem_start + TANGO2_SYSTEMRAM_ACTUALSIZE) & 0xfff00000) - em8xxx_kmem_start;
 
	/* Re-program CPU_remap so we can see full 256MB space in KSEG0/KSEG1 */
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_remap, 0x1fc00000);
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_remap1, 0x00000000);
	iob();

#ifdef CONFIG_UART_USE_SYSCLK
 	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_UART0_base + CPU_UART_CLKSEL, 0);
 	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_UART1_base + CPU_UART_CLKSEL, 0);
#else
 	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_UART0_base + CPU_UART_CLKSEL, 1);
 	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_UART1_base + CPU_UART_CLKSEL, 1);
#endif

	/* you should these macros defined in include/asm/bootinfo.h */
        mips_machgroup = MACH_GROUP_SIGMADESIGNS;
        mips_machtype = MACH_TANGO2;
	__write_32bit_c0_register($15, 1, kernel_entry);

	kend = KSEG1ADDR(em8xxx_kmem_start + em8xxx_kmem_size) - KSEG1ADDR(MEM_BASE_dram_controller_0);

	if (is_valid_memcfg(m) == 0) {
		memset(m,0,sizeof(memcfg_t));
		m->signature=MEMCFG_SIGNATURE;
		m->dram0_size=TANGO2_SYSTEMRAM_ACTUALSIZE;
		m->kernel_end=kend;
		gen_memcfg_checksum(m);
//		emprintk("ebase @%08lx. memcfg ready @%08lx.\n", read_c0_ebase(), MEM_BASE_dram_controller_0+FM_MEMCFG);
	} else {
		m->kernel_end = kend;
		gen_memcfg_checksum(m);
//		emprintk("ebase @%08lx. found memcfg @%08lx.\n", read_c0_ebase(), MEM_BASE_dram_controller_0+FM_MEMCFG);
	}

#ifdef CONFIG_QUICKTURN
	// emhack: removing 16MB here!
	add_memory_region(MEM_BASE_dram_controller_0+16*1024*1024, TANGO2_SYSTEMRAM_ACTUALSIZE, BOOT_MEM_RAM);
#else
	offset = KSEG1ADDR(em8xxx_kmem_start) - KSEG1ADDR(MEM_BASE_dram_controller_0);
	/* Mark the address from 0 to the start of kernel reserved */
	add_memory_region(0, MEM_BASE_dram_controller_0+offset, BOOT_MEM_RESERVED);
	add_memory_region(MEM_BASE_dram_controller_0+offset, em8xxx_kmem_size, BOOT_MEM_RAM);
                                                                                
//	emprintk("Adding memory region from 0x%08lx, size 0x%08lx\n", MEM_BASE_dram_controller_0+offset, em8xxx_kmem_size);
#endif

//	emprintk("status=%p ebase=%p watchlo=%p watchhi=%p\n",read_c0_status(),read_c0_ebase(),read_c0_watchlo0(),read_c0_watchhi0());

#ifdef CONFIG_TANGO2_SMP863X
	/* Calculating CPU/System frequencies */
	em8xxx_cpu_frequency = tango2_get_cpuclock();
	em8xxx_sys_frequency = tango2_get_sysclock();
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
	tango2_device_info();
#endif
	return(0);
}

void __init prom_free_prom_memory(void)
{
}

void __init prom_fixup_mem_map(unsigned long start, unsigned long end)
{
}

#if defined(CONFIG_TANGO2_SIG_BLOCK)
int tango2_pcidev_irq_map(const int pci_idsel, const int int_num)
{
	int irq = pcidev_intr_num(sigptr, pci_idsel, int_num);
	if (irq >= 0)
		irq += IRQ_CONTROLLER_IRQ_BASE;
	return(irq);
}

int tango2_isaide_irq_map(void)
{
	int irq = isaide_intr_num(sigptr);
	if (irq >= 0)
		irq += IRQ_CONTROLLER_IRQ_BASE;
	return(irq);
}

int tango2_isaide_cs_select(void)
{
	return(isaide_cs_select(sigptr));
}

int tango2_isaide_timing_slot(void)
{
	return(isaide_timing_slot(sigptr));
}

int tango2_isaide_enabled(void)
{
	return(isaide_enabled(sigptr));
}

int tango2_bmide_enabled(void)
{
	return(bmide_enabled(sigptr));
}

int tango2_sata_enabled(void)
{
	return(sata_enabled(sigptr));
}

int tango2_pci_host_enabled(void)
{
	return(pci_host_enabled(sigptr));
}

int tango2_ir_enabled(void)
{
	return(ir_enabled(sigptr));
}

int tango2_fip_enabled(void)
{
	return(fip_enabled(sigptr));
}

int tango2_ethernet_enabled(void)
{
	return(ethernet_enabled(sigptr));
}

int tango2_sdio_enabled(void)
{
	return(sdio_enabled(sigptr));
}

int tango2_usb_enabled(void)
{
	return(usb_enabled(sigptr));
}

int tango2_i2cm_enabled(void)
{
	return(i2cm_enabled(sigptr));
}

int tango2_i2cs_enabled(void)
{
	return(i2cs_enabled(sigptr));
}

int tango2_scard_enabled(void) 
{
	return(scard_enabled(sigptr));
}

int tango2_gnet_enabled(void) 
{
	return(gnet_enabled(sigptr));
}

int tango2_pcidev_enabled(const int pci_idsel)
{
	if (tango2_pci_host_enabled())
		return(pcidev_enabled(sigptr, pci_idsel));
	else
		return(0);
}

int tango2_ethernet_getmac(unsigned char *mac)
{
	if (tango2_ethernet_enabled() == 0)
		return(0);
	else {
		mac[0] = (sigptr->hwinfo.mac_hi >> 8) & 0xff;
		mac[1] = (sigptr->hwinfo.mac_hi >> 0) & 0xff;
		mac[2] = (sigptr->hwinfo.mac_lo >> 24) & 0xff;
		mac[3] = (sigptr->hwinfo.mac_lo >> 16) & 0xff;
		mac[4] = (sigptr->hwinfo.mac_lo >> 8) & 0xff;
		mac[5] = (sigptr->hwinfo.mac_lo >> 0) & 0xff;
	}
	return(1);
}

#elif defined(CONFIG_TANGO2_XENV)

int tango2_isaide_enabled(void)
{
	return((((xenv_enabled_devices >> ISAIDE_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_bmide_enabled(void)
{
	return((((xenv_enabled_devices >> BMIDE_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_sata_enabled(void)
{
	return((((xenv_enabled_devices >> SATA_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_pci_host_enabled(void)
{
	return((((xenv_enabled_devices >> PCIHOST_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_ir_enabled(void)
{
	return((((xenv_enabled_devices >> IR_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_fip_enabled(void)
{
	return((((xenv_enabled_devices >> FIP_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_ethernet_enabled(void)
{
	return((((xenv_enabled_devices >> ETHERNET_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_sdio_enabled(void)
{
	return((((xenv_enabled_devices >> SDIO_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_usb_enabled(void)
{
	return((((xenv_enabled_devices >> USB_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_i2cm_enabled(void)
{
	return((((xenv_enabled_devices >> I2CM_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_i2cs_enabled(void)
{
	return((((xenv_enabled_devices >> I2CS_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_scard_enabled(void) 
{
	return((((xenv_enabled_devices >> SCARD_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_gnet_enabled(void) 
{
	return((((xenv_enabled_devices >> GNET_SHIFT) & 1) != 0) ? 1 : 0);
}

int tango2_pcidev_enabled(const int pci_idsel)
{
	return((((xenv_enabled_devices >> (pci_idsel - 1 + PCI1_SHIFT)) & 1) != 0) ? 1 : 0);
}

int tango2_ethernet_getmac(unsigned char *mac)
{
	if (tango2_ethernet_enabled() == 0)
		return(0);
	else if ((xenv_mac_hi != 0) || (xenv_mac_lo != 0)) {
		mac[0] = (xenv_mac_hi >> 8) & 0xff;
		mac[1] = (xenv_mac_hi >> 0) & 0xff;
		mac[2] = (xenv_mac_lo >> 24) & 0xff;
		mac[3] = (xenv_mac_lo >> 16) & 0xff;
		mac[4] = (xenv_mac_lo >> 8) & 0xff;
		mac[5] = (xenv_mac_lo >> 0) & 0xff;
		return(1);
	}
	return(0);
}

int tango2_pcidev_irq_map(const int pci_idsel, const int int_num)
{
	int route = xenv_pcidev_irq_route[pci_idsel - 1];
	int irq;

	irq = (int)((route >> (int_num * 8)) & 0x3); /* int_num: 0-3 = INTA-D */
	if (irq >= 0)
		irq += (IRQ_CONTROLLER_IRQ_BASE + LOG2_CPU_PCI_INTA);
	return(irq);
}

int tango2_isaide_irq_map(void)
{
	int irq = 0;

	if (tango2_isaide_enabled() == 0)
		return(-1);

	if (xenv_isaide_irq >= 0)
		irq = xenv_isaide_irq + (IRQ_CONTROLLER_IRQ_BASE + LOG2_CPU_PCI_INTA);
	return(irq);
}

int tango2_isaide_cs_select(void)
{
	int i;
	unsigned long cs_config = (gbus_read_uint32(pGBus, REG_BASE_host_interface + PB_CS_config) >> 12) & 0xf;

	if (tango2_isaide_enabled() == 0)
		return(-1);

	for (i = 0; i < 4; i++) {
		if ((cs_config & 0x1) != 0)
			return(i);
		else
			cs_config >>= 1;
	}
	return(-1);
}

int tango2_isaide_timing_slot(void)
{
	return(xenv_isaide_timing_slot & 0x7);
}

#endif 

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)

static int tango2_device_info(void)
{
	int i;

#if defined(CONFIG_TANGO2_SIG_BLOCK)
	printk("SMP863x Enabled Devices under Linux/SIGBLOCK = 0x%08lx\n ", sigptr->hwinfo.dev_enabled);
#else
	printk("SMP863x Enabled Devices under Linux/XENV 0x%08lx = 0x%08lx\n ", xenv_gbus_addr, xenv_enabled_devices);
#endif
	if (tango2_isaide_enabled())
		printk(" ISA/IDE");
	if (tango2_bmide_enabled())
		printk(" BM/IDE");
	if (tango2_pci_host_enabled())
		printk(" PCIHost");
	if (tango2_ethernet_enabled())
		printk(" Ethernet");
	if (tango2_ir_enabled())
		printk(" IR");
	if (tango2_fip_enabled())
		printk(" FIP");
	if (tango2_i2cm_enabled())
		printk(" I2CM");
	if (tango2_i2cs_enabled())
		printk(" I2CS");
	if (tango2_sdio_enabled())
		printk(" SDIO");
	if (tango2_usb_enabled())
		printk(" USB");
	for (i = 1; i <= 6; i++) {
		if (tango2_pcidev_enabled(i))
			printk(" PCIDev%d", i);
	}
	if (tango2_sata_enabled())
		printk(" SATA");
	if (tango2_scard_enabled())
		printk(" SCARD");
	if (tango2_gnet_enabled())
		printk(" GNET");
	printk("\n");
}

EXPORT_SYMBOL(tango2_pcidev_irq_map);
EXPORT_SYMBOL(tango2_isaide_irq_map);
EXPORT_SYMBOL(tango2_isaide_cs_select);
EXPORT_SYMBOL(tango2_isaide_timing_slot);
EXPORT_SYMBOL(tango2_isaide_enabled);
EXPORT_SYMBOL(tango2_bmide_enabled);
EXPORT_SYMBOL(tango2_sata_enabled);
EXPORT_SYMBOL(tango2_pci_host_enabled);
EXPORT_SYMBOL(tango2_ir_enabled);
EXPORT_SYMBOL(tango2_fip_enabled);
EXPORT_SYMBOL(tango2_ethernet_enabled);
EXPORT_SYMBOL(tango2_sdio_enabled);
EXPORT_SYMBOL(tango2_usb_enabled);
EXPORT_SYMBOL(tango2_scard_enabled);
EXPORT_SYMBOL(tango2_gnet_enabled);
EXPORT_SYMBOL(tango2_i2cm_enabled);
EXPORT_SYMBOL(tango2_i2cs_enabled);
EXPORT_SYMBOL(tango2_pcidev_enabled);
EXPORT_SYMBOL(tango2_ethernet_getmac);

#if defined(CONFIG_TANGO2_XENV)
EXPORT_SYMBOL(xenv_scard_5v_pin);
EXPORT_SYMBOL(xenv_scard_off_pin);
EXPORT_SYMBOL(xenv_scard_cmd_pin);
#endif

#endif /* CONFIG_TANGO2_SIG_BLOCK || defined(CONFIG_TANGO2_XENV) */

EXPORT_SYMBOL(tango2_get_sysclock);
EXPORT_SYMBOL(tango2_get_cpuclock);

RMuint32 gbus_read_dram_uint32(struct gbus *pgbus, RMuint32 byte_address);
RMuint16 gbus_read_dram_uint16(struct gbus *pgbus, RMuint32 byte_address);
RMuint8 gbus_read_dram_uint8(struct gbus *pgbus, RMuint32 byte_address);
void gbus_write_dram_uint32(struct gbus *pgbus, RMuint32 byte_address, RMuint32 data);
void gbus_write_dram_uint16(struct gbus *pgbus, RMuint32 byte_address, RMuint16 data);
void gbus_write_dram_uint8(struct gbus *pgbus, RMuint32 byte_address, RMuint8 data);

#define TMP_REMAPPED_REG   CPU_remap4
#define TMP_REMAPPED_BASE  CPU_remap4_address

static RMuint32 set_remap(RMuint32 remap_reg, RMuint32 value)
{
	RMuint32 orig = *((volatile RMuint32 *)KSEG1ADDR(REG_BASE_cpu_block + remap_reg));
	if (orig != value) {
		*((volatile RMuint32 *)KSEG1ADDR(REG_BASE_cpu_block + remap_reg)) = value;
		iob();
	}
	return(orig);
}

RMuint32 gbus_read_uint32(struct gbus *pgbus, RMuint32 byte_address)
{
	RMuint32 remap;
	RMuint32 tmp;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		return gbus_read_dram_uint32(pgbus, byte_address);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			return gbus_read_dram_uint32(pgbus, byte_address);
		else {
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
			return(0);
		}
#else
		return gbus_read_dram_uint32(pgbus, byte_address);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		tmp = *((volatile RMuint32 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff)));
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
		return(tmp);
	}
}

RMuint16 gbus_read_uint16(struct gbus *pgbus, RMuint32 byte_address)
{
	RMuint32 remap;
	RMuint16 tmp;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		return gbus_read_dram_uint16(pgbus, byte_address);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			return gbus_read_dram_uint16(pgbus, byte_address);
		else {
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
			return(0);
		}
#else
		return gbus_read_dram_uint16(pgbus, byte_address);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		tmp = *((volatile RMuint16 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff)));
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
		return(tmp & 0xffff);
	}
}

RMuint8 gbus_read_uint8(struct gbus *pgbus, RMuint32 byte_address)
{
	RMuint32 remap;
	RMuint8 tmp;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		return gbus_read_dram_uint8(pgbus, byte_address);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			return gbus_read_dram_uint8(pgbus, byte_address);
		else {
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
			return(0);
		}
#else
		return gbus_read_dram_uint8(pgbus, byte_address);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		tmp = *((volatile RMuint8 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff)));
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
		return(tmp & 0xff);
	}
}

void gbus_write_uint32(struct gbus *pgbus, RMuint32 byte_address, RMuint32 data)
{
	RMuint32 remap;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		gbus_write_dram_uint32(pgbus, byte_address, data);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			gbus_write_dram_uint32(pgbus, byte_address, data);
		else 
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
#else
		gbus_write_dram_uint32(pgbus, byte_address, data);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		*((volatile RMuint32 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff))) = data;
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
	}
	__sync();
}

void gbus_write_uint16(struct gbus *pgbus, RMuint32 byte_address, RMuint16 data)
{
	RMuint32 remap;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		gbus_write_dram_uint16(pgbus, byte_address, data);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			gbus_write_dram_uint16(pgbus, byte_address, data);
		else 
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
#else
		gbus_write_dram_uint16(pgbus, byte_address, data);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		*((volatile RMuint16 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff))) = (data & 0xffff);
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
	}
	__sync();
}

void gbus_write_uint8(struct gbus *pgbus, RMuint32 byte_address, RMuint8 data)
{
	RMuint32 remap;

	if (byte_address < (MEM_BASE_dram_controller_0 + 0x10000000))
		gbus_write_dram_uint8(pgbus, byte_address, data);
	else if (byte_address < (MEM_BASE_dram_controller_1 + 0x10000000))
#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1 
		if (byte_address < (MEM_BASE_dram_controller_1 + em86xx_tlb_dram1_map_size))
			gbus_write_dram_uint8(pgbus, byte_address, data);
		else 
			printk("accessing non-existed DRAM1 area 0x%08lx.\n", byte_address);
#else
		gbus_write_dram_uint8(pgbus, byte_address, data);
#endif
	else {
		unsigned long flags;
		local_irq_save(flags); // Ensure remap won't be changed 
		// Use CPU_remapx to temporarily map the address
		remap = set_remap(TMP_REMAPPED_REG, byte_address & 0xfc000000);
		*((volatile RMuint8 *)KSEG1ADDR(TMP_REMAPPED_BASE + (byte_address & 0x03ffffff))) = (data & 0xff);
		set_remap(TMP_REMAPPED_REG, remap);
		local_irq_restore(flags);
	}
	__sync();
}

EXPORT_SYMBOL(gbus_read_uint8);
EXPORT_SYMBOL(gbus_read_uint16);
EXPORT_SYMBOL(gbus_read_uint32);
EXPORT_SYMBOL(gbus_write_uint8);
EXPORT_SYMBOL(gbus_write_uint16);
EXPORT_SYMBOL(gbus_write_uint32);

#ifdef CONFIG_TANGO2_USE_TLB_REMAP_DRAM1
EXPORT_SYMBOL(em86xx_tlb_dram1_map_base);
EXPORT_SYMBOL(em86xx_tlb_dram1_map_size);
#endif

