/*
 * Copyright 2001 MontaVista Software Inc.
 * Author: Jun Sun, jsun@mvista.com or jsun@junsun.net
 *
 * arch/mips/tango2/setup.c
 *     The setup file for tango2.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/serial.h>
#include <linux/types.h>
#include <linux/string.h>	/* for memset */
#include <linux/ioport.h>
#include <linux/console.h>

#include <asm/reboot.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/pgtable.h>
#include <asm/processor.h>
#include <asm/reboot.h>
#include <asm/system.h>

#include <asm/tango2/tango2.h>

#include <asm/tango2/emhwlib_lram.h>

extern int em86xx_sbox_init(void);
extern void pci_disable_device_all(void);
 
void tango2_restart(char *command)
{
	unsigned long flags, tmp;

	local_irq_save(flags); /* Disable interrupts */

#ifndef CONFIG_TANGO2_USE_CPU_CLOCK
	/* Disable timer */
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_time0_ctrl, 0);
#endif
#ifdef CONFIG_PCI
	/* Disable all pci devices,
	 * for reboot problem after loading usb drivers*/
	pci_disable_device_all();
#endif

#if (EM86XX_REVISION > 3)
        /* Resetting Tango2 EHCI */
        tmp = gbus_read_uint32(pGBus, REG_BASE_host_interface + 0x1410);
        tmp &= ~1;
        gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1410, tmp);
        mdelay(5);

        /* Resetting Tango2 OHCI */
        gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1514, 1<<31);
        gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1504, 0);
        mdelay(5);

	/* Resetting internal USB PHY in USB Control space */
	tmp = gbus_read_uint32(pGBus, REG_BASE_host_interface + 0x1700);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1700, tmp | 1);
	udelay(30);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1700, tmp);
	mdelay(5);

	/* Resetting internal OHCI in USB OHCI space*/
	tmp = gbus_read_uint32(pGBus, REG_BASE_host_interface + 0x1508);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1508, tmp | 0x01);

	/* Reseting OHCI dpll, it says the bit is for simulation */
	tmp = gbus_read_uint32(pGBus, REG_BASE_host_interface + 0x1700);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x1700, tmp | (1<<19));
	mdelay(1);

	/* Resetting Tango2 Ethernet interface */
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x7018, 0);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x701c, 0);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x7000, 1);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + 0x7000, 0);
#endif

	/* Resetting Video block */
	gbus_write_uint32(pGBus, REG_BASE_display_block + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_display_block + G2L_RESET_CONTROL, 2);

	/* Resetting MPEG0 block */
	gbus_write_uint32(pGBus, REG_BASE_mpeg_engine_0 + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_mpeg_engine_0 + G2L_RESET_CONTROL, 2);

	/* Resetting MPEG1 block */
	gbus_write_uint32(pGBus, REG_BASE_mpeg_engine_1 + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_mpeg_engine_1 + G2L_RESET_CONTROL, 2);

	/* Resetting Transport demux block */
	gbus_write_uint32(pGBus, REG_BASE_demux_engine + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_demux_engine + G2L_RESET_CONTROL, 2);

	/* Resetting Audio0 block */
	gbus_write_uint32(pGBus, REG_BASE_audio_engine_0 + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_audio_engine_0 + G2L_RESET_CONTROL, 2);

	/* Resetting Audio1 block */
	gbus_write_uint32(pGBus, REG_BASE_audio_engine_1 + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_audio_engine_1 + G2L_RESET_CONTROL, 2);

	/* Resetting host interface */
	gbus_write_uint32(pGBus, REG_BASE_host_interface + G2L_RESET_CONTROL, 3);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + G2L_RESET_CONTROL, 2);

#ifdef CONFIG_TANGO2_XENV
	/* nowhere to  jump, everything is  in xload format,  lets ask
 	 * xpu to reboot */
	{
#define XRPC_ID_REBOOT		19
#define SOFT_IRQ_XRPC		(1 << 4)

		struct xrpc_block_header {
			u32 callerid;
			u32 xrpcid;
			u32 param0;
			u32 param1;
			u32 param2;
			u32 param3;
			u32 param4;
			u32 headerandblocksize;
		};

		int loop;
		unsigned long base_addr = DMEM_BASE_audio_engine_0;
		struct xrpc_block_header *pB = (struct xrpc_block_header *)base_addr;

		gbus_write_uint32(pGBus, (unsigned long)&pB->callerid, 0);
		gbus_write_uint32(pGBus, (unsigned long)&pB->headerandblocksize, 
				(sizeof(struct xrpc_block_header) + 63) & ~63);
		gbus_write_uint32(pGBus, (unsigned long)&pB->xrpcid, XRPC_ID_REBOOT);

		/* try to lock xrpc mutex for at most 1 sec */
		for (loop = 0; loop < 1000; loop++) {
			if (!gbus_read_uint32(pGBus, REG_BASE_host_interface + host_mutex10))
				break;
			mdelay(1);
		}
		gbus_write_uint32(pGBus, REG_BASE_cpu_block + LR_XPU_STAGE, (unsigned long)pB);

		/* cross our fingers now */
		gbus_write_uint32(pGBus, REG_BASE_irq_handler_block + CPU_irq_softset, SOFT_IRQ_XRPC);
		while (1);
	}
#else
	/* Now to handle CPU side */
        set_c0_status(ST0_BEV | ST0_ERL);
        change_c0_config(CONF_CM_CMASK, CONF_CM_UNCACHED);
        flush_cache_all();
        write_c0_wired(0);
        __asm__ __volatile__("jr\t%0"::"r"(0xbfc00000));
#endif
}

void tango2_halt(void)
{
        printk(KERN_NOTICE "\n** You can safely turn off the power\n");
        while (1);
}

void tango2_power_off(void)
{
        tango2_halt();
}

static void __init tango2_time_init(void)
{
//	emprintk("\n");

#ifdef CONFIG_QUICKTURN
	mips_hpt_frequency = TANGO2_CPU_FREQUENCY/2;
#elif defined(CONFIG_TANGO2_SMP863X)
	{
		extern unsigned long em8xxx_cpu_frequency;
		mips_hpt_frequency = em8xxx_cpu_frequency / 2;
	}
#else
#error "Unsupported platform"
#endif
}

// CPU_time0_load
//   clock / HZ / (2 * prescale)
// CPU_time0_ctrl
//   PS(D2-3) : prescale. 0x00 = 1, 0x01 = 16, 0x10 = 256
//     There is a bug, and the actual prescale is 0x01 = 32, 0x10 = 512
//   M(D6) : periodic mode
//   E(D7) : enable

extern unsigned long em8xxx_sys_frequency;

#define TICKS_PER_SEC           em8xxx_sys_frequency
#define TIMER_PRESCALE          512 
#define TIMER_PRESCALEBITS      9
#define TIMER_RELOAD            ((TICKS_PER_SEC / HZ) >> (TIMER_PRESCALEBITS))

#define TIMER_ENABLE            0x80    // D7
#define TIMER_PERIODIC          0x40    // D6
#define TIMER_PRESCALE_1        0x00    // D[2-3] = 00b
#define TIMER_PRESCALE_32       0x04    // D[2-3] = 01b
#define TIMER_PRESCALE_512      0x08    // D[2-3] = 10b

static void __init tango2_timer_setup(struct irqaction *irq)
{
//	emprintk("\n");

#ifdef CONFIG_TANGO2_USE_CPU_CLOCK
      	setup_irq(MIPS_CPU_IRQ_BASE+STATUSB_IP7-STATUSB_IP0, irq);
#else
	// CPU_time0_load register contains just 16-bits value
	// So, take cate not the value to overflow
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_time0_load, TIMER_RELOAD);
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_time0_ctrl, TIMER_ENABLE | TIMER_PERIODIC | TIMER_PRESCALE_512);
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_time0_clr, 1);

	/* Setup Timer0 as the source */
	setup_irq(LOG2_CPU_TIMER0_INT + IRQ_CONTROLLER_IRQ_BASE, irq);
#endif
	/* Clear heart beat counter */
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + LR_HB_CPU, 0);
}


void __init tango2_setup(void)
{
//	emprintk("\n");

#ifdef CONFIG_VT
#if defined(CONFIG_DUMMY_CONSOLE)
	conswitchp = &dummy_con;
#endif
#endif

        board_time_init = tango2_time_init;
        board_timer_setup = tango2_timer_setup;

        _machine_restart = tango2_restart;
        _machine_halt = tango2_halt;
        _machine_power_off = tango2_power_off;

	printk("SMP863x Chip (Configured: REVISION=0x%x, Detected: 0x%x/0x%x)\n", 
			EM86XX_REVISION, 
			gbus_read_uint32(pGBus, REG_BASE_host_interface + PCI_REG0) & 0xffff,
			gbus_read_uint32(pGBus, REG_BASE_host_interface + PCI_REG1) & 0xff);

	/* I/O regions */
	ioport_resource.start = iomem_resource.start = 0x0;
	ioport_resource.end = 0x80000000UL - 1;
	iomem_resource.end = 0x80000000UL - 1;

  	/* Initialize Switch box */
  	em86xx_sbox_init();

#ifdef CONFIG_TANGO2_HACK
 	/* This portion will be done in bootloader later */
 
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_default_timing, 0x10101010);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_CS_config, 0x00001044);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing0, 0x10101010);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing0, 0x000001f4);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing1, 0x00110101);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing1, 0x000003f3);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing2, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing2, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing3, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing3, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing4, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing4, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing5, 0);
 	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_use_timing5, 0);
                                                                                 
 	// use GPIO#8, IRQ 14 for PCI IRQ, ISA IDE uses GPIO#6.
 	gbus_write_uint32(pGBus, REG_BASE_system_block + SYS_gpio_int, 0x0607080d);
 
	// Set the PCI IRQ to be active low, level triggered
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_edge_config_rise_set, 1 << LOG2_CPU_PCI_INTB);
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_edge_config_fall_set, 1 << LOG2_CPU_PCI_INTB);
#endif
}

