
/*****************************************
 Copyright © 2001-2003  
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file  tango2controllers.c
  @brief  

  @author Emmanuel Michon
  @date   2005-01-20
*/

#include "tango2controllers.h"

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/signal.h>

#include <asm/irq_cpu.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/ptrace.h>

#include <asm/tango2/tango2.h>


static unsigned long tango2_irq_bases[3];

#if (EM86XX_REVISION > 3)
static unsigned long disabled_while_in_progress[6]={};
#else
static unsigned long disabled_while_in_progress[3]={};
#endif

static inline void tango2_irq_enable(unsigned int x) 
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - IRQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIRQ;
	
#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableset_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableset,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableset,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#endif
}	

static inline void tango2_fiq_enable(unsigned int x) 
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - FIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLFIQ;
	
#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableset_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableset,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableset,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#endif
}	

static inline void tango2_iiq_enable(unsigned int x) 
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - IIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIIQ;
	
#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableset_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableset,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableset,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]&=~(1<<bit);
#endif
}	

static void tango2_irq_disable(unsigned int x)
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - IRQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIRQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableclr_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableclr,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_irq_enableclr,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#endif
}

static void tango2_fiq_disable(unsigned int x)
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - FIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLFIQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableclr_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableclr,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_fiq_enableclr,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#endif
}

static void tango2_iiq_disable(unsigned int x)
{
	irq_desc_t *desc = irq_desc + x;
	int bit = x - IIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIIQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		bit -= 32;
		index += 3;
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableclr_hi,1<<bit);
	} else {
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableclr,1<<bit);
	}
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_iiq_enableclr,1<<bit);
	if (desc->status&IRQ_INPROGRESS)
		disabled_while_in_progress[index]|=(1<<bit);
#endif
}

static unsigned int tango2_irq_startup(unsigned int x)
{ 
	int bit = x - IRQ_CONTROLLER_IRQ_BASE;
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
	tango2_irq_enable(x); 
	return 0; 
}

#define	tango2_irq_shutdown tango2_irq_disable

static unsigned int tango2_fiq_startup(unsigned int x)
{ 
	int bit = x - FIQ_CONTROLLER_IRQ_BASE;
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
	tango2_fiq_enable(x); 
	return 0; 
}

#define	tango2_fiq_shutdown tango2_fiq_disable

static unsigned int tango2_iiq_startup(unsigned int x)
{ 
	int bit = x - IIQ_CONTROLLER_IRQ_BASE;
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
	tango2_iiq_enable(x); 
	return 0; 
}

#define	tango2_iiq_shutdown tango2_iiq_disable

static void tango2_irq_ack(unsigned int x) 
{
	int bit = x - IRQ_CONTROLLER_IRQ_BASE;

	// we are _not_ INPROGRESS
	tango2_irq_disable(x);
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
}

static void tango2_fiq_ack(unsigned int x) 
{
	int bit = x - FIQ_CONTROLLER_IRQ_BASE;

	// we are _not_ INPROGRESS
	tango2_fiq_disable(x);
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
}

static void tango2_iiq_ack(unsigned int x) 
{
	int bit = x - IIQ_CONTROLLER_IRQ_BASE;

	// we are _not_ INPROGRESS
	tango2_iiq_disable(x);
#if (EM86XX_REVISION > 3)
	if (bit >= 32) 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat_hi,1<<(bit-32));
	else 
		gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#else
	gbus_write_uint32(pGBus,REG_BASE_cpu_block+CPU_edge_rawstat,1<<bit);
#endif
}

/* Re-enabling interrupt in the irq_end operations */
static void tango2_irq_end(unsigned int x) 
{ 
	int bit = x - IRQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIRQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		index += 3;
		bit -= 32;
	}
#endif
	if (!(disabled_while_in_progress[index]&(1<<bit)))
		tango2_irq_enable(x);
}

/* Re-enabling interrupt in the fiq_end operations */
static void tango2_fiq_end(unsigned int x) 
{ 
	int bit = x - FIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLFIQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		index += 3;
		bit -= 32;
	}
#endif
	if (!(disabled_while_in_progress[index]&(1<<bit)))
		tango2_fiq_enable(x);
}

/* Re-enabling interrupt in the iiq_end operations */
static void tango2_iiq_end(unsigned int x) 
{ 
	int bit = x - IIQ_CONTROLLER_IRQ_BASE;
	int index = TANGO2_CTRLIIQ;

#if (EM86XX_REVISION > 3)
	if (bit >= 32) {
		index += 3;
		bit -= 32;
	}
#endif
	if (!(disabled_while_in_progress[index]&(1<<bit)))
		tango2_iiq_enable(x);
}

static hw_irq_controller tango2_irq_controller[3] = {
	{	
		"tango2_irq",
		tango2_irq_startup,
		tango2_irq_shutdown,
		tango2_irq_enable,
		tango2_irq_disable,
		tango2_irq_ack,
		tango2_irq_end,
		NULL			/* no affinity stuff for UP */
	},
	{	
		"tango2_fiq",
		tango2_fiq_startup,
		tango2_fiq_shutdown,
		tango2_fiq_enable,
		tango2_fiq_disable,
		tango2_fiq_ack,
		tango2_fiq_end,
		NULL			/* no affinity stuff for UP */
	},
	{	
		"tango2_iiq",
		tango2_iiq_startup,
		tango2_iiq_shutdown,
		tango2_iiq_enable,
		tango2_iiq_disable,
		tango2_iiq_ack,
		tango2_iiq_end,
		NULL			/* no affinity stuff for UP */
	},
};

void tango2_irq_init(int ctrlno,int irq_base)
{
	int x;

#if (EM86XX_REVISION > 3)
	/* SMP863xES4 or above has 64 interrupt sources */
	for (x = irq_base; x < irq_base + 64; x++) {
		irq_desc[x].status = IRQ_DISABLED;
		irq_desc[x].action = NULL;
		irq_desc[x].depth = 1;
		irq_desc[x].handler = &tango2_irq_controller[ctrlno];
	}
#else
	/* SMP8630ES1-3 has 32 interrupt sources */
	for (x = irq_base; x < irq_base + 32; x++) {
		irq_desc[x].status = IRQ_DISABLED;
		irq_desc[x].action = NULL;
		irq_desc[x].depth = 1;
		irq_desc[x].handler = &tango2_irq_controller[ctrlno];
	}
#endif

	tango2_irq_bases[ctrlno] = irq_base;
}

