/*****************************************
 Copyright  2004-2005
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/

#ifdef __KERNEL__
#include <linux/config.h>
#include <linux/string.h>
#else
#include "zbootcfg.h"
#include "uart.h"
#include "util.h"
#endif

#if defined(CONFIG_XENV_SUPPORT) || defined(CONFIG_TANGO2_XENV)

#include "xenv.h"

// #include "xos.h"

#define XENV_RDONLY

#define xos_strcmp strcmp
#define xos_strlen strlen
#define xos_memcpy memcpy
#define RMuint32_memcmp memcmp

int xenv_isvalid(RMuint32 *base,RMuint32 size);
void full_sha1(RMuint32 digest[5],const RMuint8 *src,RMuint32 size,RMuint32 fromISP);

#if 0
#define LOCALDBG ENABLE
#else
#define LOCALDBG DISABLE
#endif

/* xenv_lookup looks for a record
 *  return -1 if not found
 *  return -2 if parsing failed
 *  return offset of record if found
 */
static int xenv_lookup(RMuint32 *base,RMuint32 size,RMascii *recordname)
{
	RMascii *p;
	int i;
	int env_size;
	
	env_size=xenv_isvalid(base,size);
	
	if (env_size<0) 
		return -2;
	
	// RMDBGLOG((LOCALDBG, "[%s]\n",recordname));
	
	p=(RMascii *)base;
	i=24; 			// jump over header
	
	while(i<env_size){
		RMuint16 rec_size=((p[i]&0xf)<<8) + (((RMuint16)p[i+1])&0xff);

		if (!xos_strcmp(recordname, p+i+2)) 
			return i;
		
		i+=rec_size;
	}
	
	return -1;
}

void xenv_format(RMuint32 *base)
{
	base[0]=24;
	full_sha1(base+1,(const RMuint8 *)(base+6),24-24,0);

	// RMDBGLOG((LOCALDBG, "%p done\n", base));
}

int xenv_isvalid(RMuint32 *base,RMuint32 size)
{
	RMuint32 env_size=base[0];
	
	if ((24<=env_size)&&(env_size<=size)) {
		RMuint32 hash[5];
		
		full_sha1(hash,(const RMuint8 *)(base+6),env_size-24,0);
		if (RMuint32_memcmp(base+1,hash,5)!=0) {
			// RMDBGLOG((ENABLE,"corrupted\n"));
			return -1;
		}
	}
	else {
		// RMDBGLOG((ENABLE,"runaway %d\n",env_size));
		return -1;
	}

	return env_size;
}

RMstatus xenv_get(RMuint32 *base, RMuint32 size, RMascii *recordname, void *dst, RMuint32 *datasize)
{
	RMascii *p = (RMascii *)base;
	RMuint32 data_len;
	RMuint32 key_len;
	RMuint32 env_size;
	
	int i;	
	
	env_size=base[0];
	i = xenv_lookup(base, size, recordname);
	if(i==-2) 
		return RM_ERROR;
	if(i==-1) 
		return RM_NOT_FOUND;

	// else we found the record
	key_len=xos_strlen(recordname);
	data_len=((p[i] & 0xf)<<8) + (((RMuint32)p[i+1])&0xff);
	data_len-=2+key_len+1;
	
	if(data_len>*datasize) {
		*datasize=data_len;

		// RMDBGLOG((ENABLE,"cannot store result\n"));
		return RM_INSUFFICIENT_SIZE;
	}
	
	*datasize=data_len;
	xos_memcpy(dst, p+i+2+key_len+1, data_len);

	// RMDBGLOG((LOCALDBG, "found [%s], length %d\n",recordname,*datasize));

	return RM_OK;
}

#ifndef XENV_RDONLY
RMstatus xenv_set(RMuint32 *base, RMuint32 size, RMascii *recordname,void *src,RMuint8 attr,RMuint32 datasize)
{
	RMascii *p;
	RMuint32 env_size;
	RMuint32 rec_attr;
	RMuint32 rec_size;
	RMuint32 key_len;
	int i;

	// RMDBGLOG((LOCALDBG, "[%s], length %d\n", recordname,datasize));
	
	p=(RMascii *)base;
	i = xenv_lookup(base, size, recordname);
	if(i==-2) 
		return RM_ERROR;
	env_size=base[0];

	if(i>=0) {
		// RMDBGLOG((LOCALDBG, "deleting record\n"));
		
		rec_attr=p[i]>>4;
		rec_size=((p[i]&0xf)<<8) + (((RMuint32)p[i+1])&0xff);
			
		if ((rec_attr==XENV_ATTR_RO)&&src) {
			// RMDBGLOG((ENABLE,"wanna change ro record\n"));
			return RM_INVALIDMODE;
		}
		
		if (rec_attr==XENV_ATTR_OTP) {
			// RMDBGLOG((ENABLE,"wanna change/clear otp record\n"));
			return RM_INVALIDMODE;
		}

		// delete the record. Supposes memcpy is implemented increasing.
		xos_memcpy(p+i, p+i+rec_size, env_size-(i+rec_size));
		env_size-=rec_size;
	}
		
	// add the record at the end if needed.
	if(src) {
		i=env_size;
		key_len=xos_strlen(recordname);
		rec_size=2+key_len+1+datasize;

		if((i+rec_size)>=size)
			return RM_INSUFFICIENT_SIZE;
		
		p[i]=((attr&0xf)<<4) | ((rec_size>>8)&0xf);
		p[i+1]=rec_size&0xff;
		
		xos_memcpy(p+i+2, recordname, key_len+1);
		xos_memcpy(p+i+2+key_len+1, src, datasize);
		
		env_size+=rec_size;
	}
	
	base[0]=env_size;
	full_sha1(base+1,(const RMuint8 *)(base+6),env_size-24,0);

	return RM_OK;
}
#endif /* !XENV_RDONLY */

#ifdef _DEBUG
void xenv_dump(RMuint32 *base,RMuint32 size)
{
	RMascii *p;
	int i;
	int env_size;
	RMuint32 records=0;

#if 0
	return;
#endif
	
	env_size=xenv_isvalid(base,size);
	
	if (env_size<0) 
		return;
	
	// RMDBGLOG((ENABLE,"@%p\n",base));
	
	p=(RMascii *)base;
	i=24; 			// jump over header
	
	while(i<env_size){
		RMuint8 rec_attr;
		RMuint16 rec_size;
		RMascii *recordname;
		RMuint32 key_len;
		RMascii *x;
		
		rec_attr=p[i]>>4;
		rec_size=((p[i]&0xf)<<8) + (((RMuint16)p[i+1])&0xff);
		recordname=p+i+2;
		key_len=xos_strlen(recordname);
		
		// RMDBGLOG((ENABLE,"(0x%02x) [%s] =",rec_attr,recordname));
		for (x=recordname+key_len+1;x<recordname+rec_size-2;x++) RMDBGPRINT((ENABLE," %02x",*x));
		RMDBGPRINT((ENABLE," .\n"));
		
		records++;
		i+=rec_size;
	}
	
	// RMDBGLOG((ENABLE,"%d records, %d bytes\n\n",records,env_size));
}
#endif
#endif /* CONFIG_XENV_SUPPORT || CONFIG_TANGO2_XENV */

