/*
 * driver/ide/em86xx/em86xx_ide.c
 * 
 * EM86XX IDE driver common code
 *
 * by Ho Lee March 4, 2003
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/ide.h>
#include <linux/delay.h>
#include "em86xx_ide.h"

//
// data structure initialization
//

//
// IDE HWIF interface introduced by Linux 2.4.22
//

// EM86XX IDE interface uses 16 bits wide bus
// and byte write doesn't start write transaction

void em86xx_ide_OUTB(u8 data, unsigned long port)
{
	gbus_write_uint16(pGBus, port, (unsigned short)data);
}

void em86xx_ide_OUTW(u16 data, unsigned long port)
{
	gbus_write_uint16(pGBus, port, (unsigned short)data);
}

void em86xx_ide_OUTBSYNC(ide_drive_t *drive, u8 data, unsigned long port)
{
	gbus_write_uint16(pGBus, port, (unsigned short)data);
}

u8 em86xx_ide_INB(unsigned long port)
{
	return((u8)(gbus_read_uint16(pGBus, port) & 0xff));
}

u16 em86xx_ide_INW(unsigned long port)
{
	return((u16)(gbus_read_uint16(pGBus, port) & 0xffff));
}

void em86xx_ide_INSW(unsigned long port, void *addr, u32 count)
{
	while (count--) {
		*(u16 *)addr = (u16)(gbus_read_uint16(pGBus, port) & 0xffff);
		addr += 2;
	}
}

void em86xx_ide_OUTSW(unsigned long port, void *addr, u32 count)
{
	while (count--) {
		gbus_write_uint16(pGBus, port, *(u16 *)addr);
		addr += 2;
	}
}

//
//
// DMA handlers
// 

int em86xx_ide_dma_on(ide_drive_t *drive)
{
#if defined(CONFIG_BLK_DEV_TANGO2_BMIDEDMA) || defined(CONFIG_BLK_DEV_TANGO2_ISAIDEDMA)
	ide_hwif_t *hwif = HWIF(drive);

	if (hwif->ide_dma_host_on)
		hwif->ide_dma_host_on(drive);

	printk("IDE: DMA enabled for %s%s\n",
		drive->media == ide_disk ? "ATA DISK " : 
		(drive->media == ide_cdrom ? "ATAPI CDROM " : ""),
		drive->name);
	drive->using_dma = 1;
#else
	drive->using_dma = 0;
#endif
	return 0;
}

int em86xx_ide_dma_off(ide_drive_t *drive)
{
#if defined(CONFIG_BLK_DEV_TANGO2_BMIDEDMA) || defined(CONFIG_BLK_DEV_TANGO2_ISAIDEDMA)
	ide_hwif_t *hwif = HWIF(drive);

	if (hwif->ide_dma_host_off)
		hwif->ide_dma_host_off(drive);

	printk("IDE: DMA disabled for %s\n", drive->name);
#endif
	drive->using_dma = 0;
	return 0;
}

int em86xx_ide_dma_off_quietly(ide_drive_t *drive)
{
	drive->using_dma = 0;
	return 0;
}

int em86xx_ide_dma_test_irq(ide_drive_t *drive)
{
	return 1;
}

int em86xx_ide_dma_bad_drive(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_good_drive(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_count(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_verbose(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_retune(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_lostirq(ide_drive_t *drive)
{
	return 0;
}

int em86xx_ide_dma_timeout(ide_drive_t *drive)
{
	return 0;
}

//
// helper functions
//

#if defined(CONFIG_BLK_DEV_TANGO2_BMIDEDMA) || defined(CONFIG_BLK_DEV_TANGO2_ISAIDEDMA)
// clean or invalidate every buffer head in request queue
void em86xx_ide_dma_cache(struct request *rq, int read)
{
    struct buffer_head *bh = rq->bh;

    do {
#ifdef CONFIG_NONCOHERENT_IO
	    if (!is_physical((u32)bh->b_data)) {
		    if (read != 0)
			    dma_cache_inv((unsigned long)bh->b_data, bh->b_size);
		    else
			    dma_cache_wback((unsigned long)bh->b_data, bh->b_size);
	    }
#endif
    } while ((bh = bh->b_reqnext) != NULL);
}
#endif

int em86xx_ide_config_drive_for_dma(ide_drive_t *drive, int mode)
{
	ide_hwif_t *hwif = HWIF(drive);
#if defined(CONFIG_BLK_DEV_TANGO2_BMIDEDMA) || defined(CONFIG_BLK_DEV_TANGO2_ISAIDEDMA)
	int config_allows_dma = 0;
	struct hd_driveid *id = drive->id;
	int speed;

	// disable DMA for CDROM by default
	if (drive->media == ide_disk || drive->media == ide_cdrom)
		config_allows_dma = 1;

	if (id && (id->capability & 1) && hwif->autodma && config_allows_dma) {
		speed = ide_dma_speed(drive, mode);

		/* For CDROM, we do the max. MDMA mode 2 */
//		if ((speed > XFER_MW_DMA_2) && (drive->media == ide_cdrom))
//			speed = XFER_MW_DMA_2;

		if (speed >= XFER_MW_DMA_0) {
			if (hwif->speedproc)
				hwif->speedproc(drive, speed);
			ide_config_drive_speed(drive, speed);
			return hwif->ide_dma_on(drive);
		}
	}
#endif

	return hwif->ide_dma_off_quietly(drive);
}
