/*
 * driver/ide/tango2/em86xx_isa.c
 * 
 * TANGO2 ISA IDE driver
 *
 * by Ho Lee March 4, 2003
 * CD-ROM DMA implemented by Ho Lee 2003/09/18
 */

/*
 	CDROM DMA transfer
		For ATA DMA transfer, DMA handler register "handler" and sends 
		read or write command to IDE device. But for CDROM, DMA handler
		doesn't send any command to the device, and it can not register
		handler to finish the transfer. DMA handler is called in the 
		middle of transfer.
*/

#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/ide.h>
#include <asm/tango2/tango2api.h>
#include "em86xx_ide.h"

#ifdef MODULE
#include <linux/module.h>
MODULE_AUTHOR("Craig Qu");
MODULE_DESCRIPTION(" Tango2 ISA IDE driver");
MODULE_LICENSE("GPL");
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
extern int tango2_isaide_irq_map(void);
extern int tango2_isaide_enabled(void);
extern int tango2_isaide_cs_select(void);
extern int tango2_isaide_timing_slot(void);
#endif

// in ide.c
void ide_intr(int irq, void *dev_id, struct pt_regs *regs);

#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDEDMA
#define EM86XX_ISAIDE_DMA_MERGE_BH
#define EM86XX_ISAIDE_DMA_BOUNCE_BUF

static int g_sbox_port_read = -1, g_sbox_port_write = -1;
static unsigned int g_regbase_read, g_regbase_write;
static int g_irq_read, g_irq_write;

// function prototypes
static int em86xx_isaide_dma_read(ide_drive_t *drive);
static int em86xx_isaide_dma_write(ide_drive_t *drive);
static int em86xx_isaide_dma_begin(ide_drive_t *drive);
static int em86xx_isaide_dma_end(ide_drive_t *drive);
static int em86xx_isaide_dma_check(ide_drive_t *drive);
static int em86xx_isaide_dma_host_on(ide_drive_t *drive);
static int em86xx_isaide_dma_host_off(ide_drive_t *drive);
static int em86xx_isaide_dma_lostirq(ide_drive_t *drive);
static int em86xx_isaide_dma_timeout(ide_drive_t *drive);
#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDETIMING
static void em86xx_isaide_tune_chipset_drive(int drive, int speed, int verbose);
static int em86xx_isaide_tune_chipset(ide_drive_t *drive, u8 speed);
static unsigned long get_pb_ide_timing(unsigned int sysclk, int mode);
#endif

static int em86xx_isaide_rwdma(ide_drive_t *drive, int read, struct request *rq);
static void em86xx_isaide_mbus_intr(int irq, void *arg);
static int em86xx_isaide_dma_end_io(ide_drive_t *drive);
static ide_startstop_t em86xx_isaide_dma_intr(ide_drive_t *drive);

// Misc
static struct request *crq = NULL;
static int crq_ntransfer = 0;

#ifdef EM86XX_ISAIDE_DMA_BOUNCE_BUF
#define EM86XX_ISAIDE_DMA_BOUNCE_BUF_ORDER		(19 - PAGE_SHIFT)	// maximum = 128 sectors = 2^7 sectors = 2^7 * 2^9 = 2^16
static unsigned char *g_bounce_buf = NULL;
#endif
#endif

//
// data structure initialization
//

void em86xx_isaide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port, ide_ioreg_t ctrl_port, int *irq)
{
	int i;
	ide_ioreg_t reg;

	memset(hw, 0, sizeof(*hw));

	// setup io_ports
	for (i = IDE_DATA_OFFSET, reg = data_port; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 4;
	}

	hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	hw->dma = NO_DMA;

	if (irq) {
#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
		*irq = tango2_isaide_irq_map();
#else
		*irq = IRQ_IDE;
#endif
	}
}

//
// initialize IDE driver
//

int __init em86xx_isaide_init(void)
{
	int index, i;
	ide_hwif_t *hwif;
	hw_regs_t hw;
#ifdef MODULE
        ide_drive_t *drive = NULL;
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
	if (tango2_isaide_enabled() == 0)
		return 0;
	else
		em86xx_isaide_init_hwif_ports(&hw, REG_BASE_host_interface_ISAIDE(tango2_isaide_cs_select()), 0, &hw.irq);
#else
	em86xx_isaide_init_hwif_ports(&hw, REG_BASE_host_interface_ISAIDE, 0, &hw.irq);
#endif
       index = ide_register_hw(&hw, NULL);
	printk("SMP863x ISA IDE activated as ide%d.\n", index);

	// initialize ide_hwifs data structure
	// supports only one interface
	for (i = index; i < index + 1; ++i) {
		// register DMA handler
		hwif = &ide_hwifs[i];
		hwif->OUTB = em86xx_ide_OUTB;
		hwif->OUTW = em86xx_ide_OUTW;
		hwif->OUTBSYNC = em86xx_ide_OUTBSYNC;
		hwif->INB = em86xx_ide_INB;
		hwif->INW = em86xx_ide_INW;
		hwif->INSW = em86xx_ide_INSW;
		hwif->OUTSW = em86xx_ide_OUTSW;
		hwif->rqsize = 16;
	    if (!noautodma) {
#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDEDMA
		hwif->autodma = 1;
		hwif->atapi_dma = 1;
		hwif->ultra_mask = 0x00;	// disable all Ultra DMA
		hwif->mwdma_mask = 0x07;	// enable all Multi-word DMA
		hwif->swdma_mask = 0x80;	// disable all Single-word DMA

		hwif->ide_dma_read = em86xx_isaide_dma_read;
		hwif->ide_dma_write = em86xx_isaide_dma_write;
		hwif->ide_dma_begin = em86xx_isaide_dma_begin;
		hwif->ide_dma_end = em86xx_isaide_dma_end;
		hwif->ide_dma_check = em86xx_isaide_dma_check;
		hwif->ide_dma_on = em86xx_ide_dma_on;
		hwif->ide_dma_off = em86xx_ide_dma_off;
		hwif->ide_dma_off_quietly = em86xx_ide_dma_off_quietly;
		hwif->ide_dma_test_irq = em86xx_ide_dma_test_irq;
		hwif->ide_dma_host_on = em86xx_isaide_dma_host_on;
		hwif->ide_dma_host_off = em86xx_isaide_dma_host_off;
		hwif->ide_dma_bad_drive = em86xx_ide_dma_bad_drive;
		hwif->ide_dma_good_drive = em86xx_ide_dma_good_drive;
		hwif->ide_dma_count = em86xx_ide_dma_count;
		hwif->ide_dma_verbose = em86xx_ide_dma_verbose;
		hwif->ide_dma_retune = em86xx_ide_dma_retune;
		hwif->ide_dma_lostirq = em86xx_isaide_dma_lostirq;
		hwif->ide_dma_timeout = em86xx_isaide_dma_timeout;
#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDETIMING
		hwif->speedproc = em86xx_isaide_tune_chipset;
#endif
#endif
	    }
	}

	if (!noautodma) {
#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDEDMA
#ifdef EM86XX_ISAIDE_DMA_BOUNCE_BUF
	if ((g_bounce_buf = (unsigned char *) __get_free_pages(GFP_KERNEL, EM86XX_ISAIDE_DMA_BOUNCE_BUF_ORDER)) == NULL)
		printk("EM86XX IDE : Can not allocate buffer for IDE transfer\n");
#ifdef EM86XX_BOUNCE_BUFFER_NOCACHE
	g_bounce_buf = (unsigned char *)KSEG1ADDR((unsigned long)g_bounce_buf);
	printk("Bounce buffer starts at 0x%08lx\n", (unsigned long)g_bounce_buf);
#endif
#endif	// BOUNCE_BUF
#endif	// EM86XX_ISAIDEDMA
	}
#if defined(MODULE) && defined(CONFIG_BLK_DEV_IDE)
#ifdef HWIF_PROBE_CLASSIC_METHOD
        probe_hwif(hwif);
        hwif_init(hwif);
#else
        probe_hwif_init(hwif);
#endif /* HWIF_PROBE_CLASSIC_METHOD */

        for (i = 0; i < MAX_DRIVES; ++i) {
                drive = &hwif->drives[i];
                if (drive->present && !drive->dead)
                        ide_attach_drive(drive);
        }
#endif
	return 0;
}

//
// IDE DMA processing
//

#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDEDMA

int em86xx_isaide_dma_read(ide_drive_t *drive)
{
	struct request *rq;

	drive->waiting_for_dma = 1;
	rq = HWGROUP(drive)->rq;
	// printk("em86xx_isaide_dma_read : nsect = %ld\n", rq->nr_sectors);
	return em86xx_isaide_rwdma(drive, 1, rq);
}

int em86xx_isaide_dma_write(ide_drive_t *drive)
{
	struct request *rq;

	drive->waiting_for_dma = 1;
	rq = HWGROUP(drive)->rq;
	// printk("em86xx_isaide_dma_write : nsect = %ld\n", rq->nr_sectors);
	return em86xx_isaide_rwdma(drive, 0, rq);
}

int em86xx_isaide_dma_begin(ide_drive_t *drive)
{
	return(0);
}

int em86xx_isaide_dma_end(ide_drive_t *drive)
{
	int status;

	drive->waiting_for_dma = 0;
	status = em86xx_isaide_dma_end_io(drive);
	return(status);
}

int em86xx_isaide_dma_lostirq(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int regbase = ((rq->cmd == READ) ? g_regbase_read : g_regbase_write);

	if (em86xx_mbus_inuse(regbase) != 0) 
		em86xx_mbus_reset(regbase, SBOX_IDEFLASH);
	return 0;
}

int em86xx_isaide_dma_timeout(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int regbase = ((rq->cmd == READ) ? g_regbase_read : g_regbase_write);

	if (em86xx_mbus_inuse(regbase) != 0) 
		em86xx_mbus_reset(regbase, SBOX_IDEFLASH);
	return 0;
}

int em86xx_isaide_dma_check(ide_drive_t *drive)
{
	return em86xx_ide_config_drive_for_dma(drive, 0);
}

int em86xx_isaide_dma_host_on(ide_drive_t *drive)
{
	return 0;
}

int em86xx_isaide_dma_host_off(ide_drive_t *drive)
{
	return 0;
}

#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDETIMING

void em86xx_isaide_tune_chipset_drive(int drive, int speed, int verbose)
{
	extern unsigned long em8xxx_sys_frequency;
	int mode = speed & 0x0f;
	unsigned int timing;
	
	if (speed >= XFER_PIO_0 && speed <= XFER_PIO_4)
		mode -= 8;
	else if (speed >= XFER_MW_DMA_0 && speed <= XFER_MW_DMA_2)
		mode += 5;
	else
		return;	

	// get_pb_ide_timing() receives mode in reversed order
	//   PIO : 7-3
	//   Multi-word DMA : 2-0
	mode = 7 - mode;
	timing = get_pb_ide_timing(em8xxx_sys_frequency * 11 / 10, mode);

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing_slot(tango2_isaide_timing_slot()), timing);
#else
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_timing_slot(PB_TIMINGSLOT_IDE), timing);
#endif
}

int em86xx_isaide_tune_chipset(ide_drive_t *drive, u8 speed)
{
	em86xx_isaide_tune_chipset_drive(drive->select.b.unit, speed, 1);

	return ide_config_drive_speed(drive, speed);
}

#endif

#ifdef EM86XX_ISAIDE_DMA_MERGE_BH
#define MBUS_LINEAR_MAX 	(0x2000 - 1)
#endif

/* Get the I/O size of given request */
static inline int get_request_size(struct request *rq)
{
	unsigned int rq_size;
	struct buffer_head *bh;

	for (rq_size = 0, bh = rq->bh; bh != NULL; bh = bh->b_reqnext) 
		rq_size += bh->b_size;

	return(rq_size);
}

#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDEDMA
int em86xx_isaide_request_fragment(struct request *rq)
{
	int ntransfer = 0;
	struct buffer_head *bh = rq->bh;

#ifdef EM86XX_ISAIDE_DMA_MERGE_BH
	unsigned long addr;
	unsigned int size;
	unsigned int sector;

	if (rq == crq)
		return(crq_ntransfer);

	/* Try to merge multiple BH if possible, one fragment only */
	for (ntransfer = 0, bh = rq->bh;;) {
		if (bh == NULL)
			break;
		else {
			if (++ntransfer > 1) /* More than one fragment found */
				break; 
		}
		addr = (unsigned long)bh->b_data;
		size = bh->b_size;
		sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((addr + size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else {	/* Merge the next BH */
				size += bh->b_size;
			}
		}	
	}
	if (ntransfer < 2) /* This can be done with one fragment (linear or rectangular) */
		goto done;

	/* Try to merge multiple BH if possible */
	for (ntransfer = 0, bh = rq->bh;;) {
		if (bh == NULL)
			break;
		else
			++ntransfer;
		addr = (unsigned long)bh->b_data;
		size = bh->b_size;
		sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((addr + size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else if ((size + bh->b_size) > MBUS_LINEAR_MAX)
				break;	/* Size too large */
			else {	/* Merge the next BH */
				size += bh->b_size;
			}
		}	
	}
#else
	if (rq == crq)
		return(crq_ntransfer);

	do {
		++ntransfer;
	} while ((bh = bh->b_reqnext) != NULL);
#endif

done:
	return ntransfer;
}

struct buffer_head *em86xx_isaide_dma_setup(int read, struct request *rq, int stop_if_fragment)
{
	struct buffer_head *bh = rq->bh;
#ifdef EM86XX_ISAIDE_DMA_MERGE_BH
	struct {
		unsigned long addr;
		unsigned int size;
		unsigned int sector;
	} transfer[2], *ptransfer;
	unsigned int ntransfer, size;

	crq = rq;
	/* Try to merge multiple BH if possible, one fragment only */
	for (ntransfer = size = 0, ptransfer = &transfer[0], bh = rq->bh;; ptransfer++) {
		if (bh == NULL)
			break;
		else if (++ntransfer > 1)
			break;

		ptransfer->addr = (unsigned long)bh->b_data;
		ptransfer->size = bh->b_size;
		ptransfer->sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((ptransfer->addr + ptransfer->size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else {	/* Merge the next BH */
				ptransfer->size += bh->b_size;
			}
		}	
		size += ptransfer->size;
	}
	if (ntransfer < 2) /* This can be done with one fragment (linear or rectangular) */
		goto start_transfer;

	/* Try to merge multiple BH if possible */
	for (ntransfer = size = 0, ptransfer = &transfer[0], bh = rq->bh;; ptransfer++) {
		if (bh == NULL)
			break;
		else if (++ntransfer > 2)
			break;

		ptransfer->addr = (unsigned long)bh->b_data;
		ptransfer->size = bh->b_size;
		ptransfer->sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((ptransfer->addr + ptransfer->size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else if ((ptransfer->size + bh->b_size) > MBUS_LINEAR_MAX)
				break;	/* Size too large */
			else {	/* Merge the next BH */
				ptransfer->size += bh->b_size;
			}
		}	
		size += ptransfer->size;
	}

	if ((ntransfer > 2) && (stop_if_fragment != 0)) {
		crq_ntransfer = ntransfer;
		return bh;
	}
#else
	struct buffer_head *prevbh;
	struct {
		unsigned int addr;
		unsigned int size;
	} transfer[2], *ptransfer;
	unsigned int ntransfer, size;

	crq = rq;
	for (prevbh = NULL, ntransfer = 0, ptransfer = transfer, size = 0; bh != NULL; prevbh = bh, bh = bh->b_reqnext) {
		if (++ntransfer > 2) {
			if (stop_if_fragment) {
				crq_ntransfer = ntransfer;
				return bh;
			}
			break;
		}
		ptransfer->addr = (unsigned int) bh->b_data;
		ptransfer->size = bh->b_size;
		// printk("MBUS : bh = %p, bh->next = %p, %08x, %08x\n", bh, bh->b_reqnext, ptransfer->addr, ptransfer->size);
		++ptransfer;
		size += bh->b_size;
	}
#endif

start_transfer:
	// Use DMA mode transfer
	// Transfer starts when the IDE device is ready and asserts DMARQ signal high
	crq_ntransfer = ntransfer;

	// setup PB automode registers
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_automode_start_address, 0);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_automode_control, 0x00140000 | ((read ? 1 : 0) << 16) | (size >> 1));

	// setup MBUS
	if (ntransfer == 1) {
#ifdef CONFIG_NONCOHERENT_IO
		if (!is_physical(transfer[0].addr)) {
			if (read) 
				dma_cache_inv(transfer[0].addr, transfer[0].size);
			else
				dma_cache_wback_inv(transfer[0].addr, transfer[0].size);
		}
#endif
		if (transfer[0].size <= MBUS_LINEAR_MAX)
			/* Do linear transfer */
			em86xx_mbus_setup_dma_linear(read ? g_regbase_read : g_regbase_write, SBOX_IDEFLASH,
					is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), 
					transfer[0].size);
		else {
			/* Do rectangular transfer */
			unsigned int i, horiz, lines, sz;
			for (i = 0, horiz = 1, sz = transfer[0].size; (i < 12) && ((sz & 0x01) == 0); ++i, horiz <<= 1, sz >>= 1) 
				;
			lines = transfer[0].size >> i;
			em86xx_mbus_setup_dma_rectangle(read ? g_regbase_read : g_regbase_write, SBOX_IDEFLASH,
					is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), 
					horiz, lines, horiz);
		}
	} else {
#ifdef CONFIG_NONCOHERENT_IO
		if (read) {
			if (!is_physical(transfer[0].addr)) 
				dma_cache_inv(transfer[0].addr, transfer[0].size);
			if (!is_physical(transfer[1].addr)) 
				dma_cache_inv(transfer[1].addr, transfer[1].size);
		} else {
			if (!is_physical(transfer[0].addr)) 
				dma_cache_wback_inv(transfer[0].addr, transfer[0].size);
			if (!is_physical(transfer[1].addr)) 
				dma_cache_wback_inv(transfer[1].addr, transfer[1].size);
		}
#endif
		em86xx_mbus_setup_dma_double(read ? g_regbase_read : g_regbase_write, SBOX_IDEFLASH,
			is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), transfer[0].size,
			is_physical(transfer[1].addr) ? transfer[1].addr : PHYSADDR(transfer[1].addr), transfer[1].size);
	}

	return bh;
}

int em86xx_isaide_rwdma(ide_drive_t *drive, int read, struct request *rq)
{
	int retry;
	unsigned long cj;
	static const int max_retry = 5;

	// setup switchbox and wait for its readiness
	if (g_sbox_port_read < 0) {
		for (retry = 0; ((g_sbox_port_read = em86xx_mbus_alloc_dma(SBOX_IDEFLASH, 1, &g_regbase_read, &g_irq_read)) < 0) && (retry < max_retry); retry++) {
			for (cj = jiffies + (HZ / 100); jiffies < cj;);
		}
		if (g_sbox_port_read < 0) { /* Failed to use DMA, use PIO instead */
			printk("%s: failed to obtain DMA channel, use PIO instead.\n", drive->name);
			return(1);
		}
	}
	if (g_sbox_port_write < 0) {
		for (retry = 0; ((g_sbox_port_write = em86xx_mbus_alloc_dma(SBOX_IDEFLASH, 0, &g_regbase_write, &g_irq_write)) < 0) && (retry < max_retry); retry++) {
			for (cj = jiffies + (HZ / 100); jiffies < cj;);
		}
		if (g_sbox_port_write < 0) { /* Failed to use DMA, use PIO instead */
			printk("%s: failed to obtain DMA channel, use PIO instead.\n", drive->name);
			em86xx_mbus_free_dma(g_sbox_port_read);
			g_sbox_port_read = -1;
			return(1);
		}
	}

	// setup handler
	if (drive->media == ide_disk)
		ide_set_handler(drive, &em86xx_isaide_dma_intr, 2 * WAIT_CMD, NULL);

#ifdef EM86XX_ISAIDE_DMA_BOUNCE_BUF
	if (em86xx_isaide_dma_setup(read, rq, 1)) {
		// copy buffer before transfer
		unsigned int buflen = get_request_size(rq);
		unsigned int buf = (unsigned int) g_bounce_buf;

		if (!read) {
			struct buffer_head *bh = rq->bh;
#if 1
			char *data;
			u32 size;
			extern int mbus_memcpy(u32 dst, u32 src, u32 size);
			
			/* Try to merge multiple BH for copying */
			while (bh != NULL) {
				data = bh->b_data;
				size = bh->b_size;
				for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
					if ((data + size) == bh->b_data)
						size += bh->b_size; /* Contiguous */
					else 
						break;
				}
				mbus_memcpy(PHYSADDR(buf), PHYSADDR(data), size);
#ifdef CONFIG_NONCOHERENT_IO
				dma_cache_inv(buf, size);
#endif
				buf += size;
			}
#else
			do {
				memcpy((void *) buf, bh->b_data, bh->b_size);
				buf += bh->b_size;
			} while ((bh = bh->b_reqnext) != NULL);
#endif
		}
 
#ifdef CONFIG_NONCOHERENT_IO
#ifndef EM86XX_BOUNCE_BUFFER_NOCACHE
		dma_cache_wback_inv((unsigned long)g_bounce_buf, buflen);
#endif
#endif
		// setup PB automode registers
		gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_automode_start_address, 0);
		gbus_write_uint32(pGBus, REG_BASE_host_interface + PB_automode_control, 0x00140000 | ((read ? 1 : 0) << 16) | (buflen >> 1));

		// setup MBUS
		em86xx_mbus_setup_dma(read ? g_regbase_read : g_regbase_write, SBOX_IDEFLASH, PHYSADDR((unsigned long)g_bounce_buf), buflen, NULL, NULL);
	} else
		em86xx_ide_dma_cache(rq, read);
#else
	// cache control
	em86xx_ide_dma_cache(rq, read);

	// setup MBUS
	em86xx_isaide_dma_setup(read, rq->bh, 0);
#endif

	em86xx_isaide_dma_begin(drive);

	if (drive->media == ide_disk) {
		// send IDE command
		HWIF(drive)->OUTB(read ? (drive->addressing ? WIN_READDMA_EXT : WIN_READDMA) : (drive->addressing ? WIN_WRITEDMA_EXT : WIN_WRITEDMA), IDE_COMMAND_REG);
	}

	return 0;
}
#endif

static void em86xx_isaide_mbus_intr(int irq, void *arg)
{
	int index = irq - (LOG2_CPU_HOST_MBUS_W0_INT + IRQ_CONTROLLER_IRQ_BASE);
	struct request *rq = (struct request *) arg;
	
	// printk("[MBUS] ");

	if (rq->bh != NULL) 
		em86xx_isaide_dma_setup((index < 2) ? 1 : 0, rq, 0);
}

int em86xx_isaide_dma_end_io(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;

	int stat = em86xx_mbus_wait(rq->cmd == READ ? g_regbase_read : g_regbase_write, SBOX_IDEFLASH);

	if (stat != 0) 
		goto done;

#ifdef EM86XX_ISAIDE_DMA_BOUNCE_BUF
	if (rq->cmd == READ) {
		if (em86xx_isaide_request_fragment(rq) > 2) {
			// copy buffer after transfer
			unsigned int buf = (unsigned int) g_bounce_buf;
			struct buffer_head *bh = rq->bh;
#ifdef CONFIG_NONCOHERENT_IO
#ifndef EM86XX_BOUNCE_BUFFER_NOCACHE
			dma_cache_inv(buf, get_request_size(rq));
#endif
#endif
#if 1
			{
				char *data;
				u32 size;
				extern int mbus_memcpy(u32 dst, u32 src, u32 size);
			
				/* Try to merge multiple BH for copying */
				while (bh != NULL) {
					data = bh->b_data;
					size = bh->b_size;
					for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
						if ((data + size) == bh->b_data)
							size += bh->b_size; /* Contiguous */
						else 
							break;
					}
					mbus_memcpy(PHYSADDR(data), PHYSADDR(buf), size);
#ifdef CONFIG_NONCOHERENT_IO	
					if (!is_physical((u32)data))
						dma_cache_inv((unsigned long)data, size);
#endif
					buf += size;
				}
			}
#else
			do {
				memcpy(bh->b_data, (void *) buf, bh->b_size);
				buf += bh->b_size;
			} while ((bh = bh->b_reqnext) != NULL);
#endif
		} else {
			em86xx_ide_dma_cache(rq, 1 /* READ */);
		}
	}
#else
	if (rq->cmd == READ) 
		em86xx_ide_dma_cache(rq, 1 /* READ */);
#endif

done:
	/* Free up MBUS channel */
	em86xx_mbus_free_dma(g_sbox_port_read);
	em86xx_mbus_free_dma(g_sbox_port_write);
	g_sbox_port_read = g_sbox_port_write = -1;
	return((stat != 0) ? 1 : 0);
}

ide_startstop_t em86xx_isaide_dma_intr(ide_drive_t *drive)
{
	int i;
	ide_hwif_t *hwif = HWIF(drive);
	struct request *rq = HWGROUP(drive)->rq;

	hwif->ide_dma_end(drive);

	for (i = rq->nr_sectors; i > 0;) {
		i -= rq->current_nr_sectors;
		ide_end_request(drive, 1);
	}

	return ide_stopped;
}

#endif

#ifdef CONFIG_BLK_DEV_TANGO2_ISAIDETIMING
/*
 * Calculate the peripheral bus timing needed for the register used by IDE
 * devices.
 */

/* 
 * The tables bellows use indices as Multi-word DMA 2, Multi-word DMA 1,
 * Multi-word DMA 0, PIO 4, PIO 3, PIO 2, PIO 1, and PIO 0 (From 0 to 7, 
 * respectively).
 */

/* For storing value pair */
typedef struct {
	unsigned char x;
	unsigned char y;
} value_pair;

/* Table used for calculate Ta and Tb */
static const value_pair tab_values[] = {
	{3,0}, {5,0}, {7,2}, {3,0}, {5,0}, {5,0}, {7,2}, {10,4}
};

/* Table used for calculate Tc */
static const value_pair tc_values[] = {
	{13,8}, {19,11}, {60,41}, {10,5}, {12,6}, {14,8}, {17,11}, {23,15}
};

/* Table used for calculate Td */
static const value_pair td_values[] = {
	{1,100}, {1,100}, {3,200}, {1,100}, {1,100}, {1,100}, {3,200}, {5,250}
};

/* 
 * Calculate the required timing for setting up peripheral bus timing register.
 *
 * Input: sysclk = system clock rate (in MHz)
 *        mode = IDE mode (0-2: Multiword DMA mode 2-0, 3-7: PIO mode 4-0)
 * Return: 0 = Error
 *         else = value for setting up the timing register
 */
unsigned long get_pb_ide_timing(unsigned int sysclk, int mode)
{
	unsigned char ta = 0;
	unsigned char tb = 0;
	unsigned char tc = 0;
	unsigned char td = 0;
	long temp = 0;
	const value_pair *val_ptr = (const value_pair *)0;

	/* The formula is only effective between 100-240MHz */
	if ((sysclk < 100) && (sysclk > 240))
		return(0);
	else if ((mode < 0) && (mode > 7)) /* Mode index is 0-7 only */
		return(0);

	/* (x, y) = tab_values[mode] */
	/* (Ta + Tb) = Temp = (ceiling((sysclk - 100) * x / 140) + y) */
	/* Ta = ceiling(Temp / 2) */
	/* Tb = (Temp / 2) */
	val_ptr = &tab_values[mode];
	temp = ((sysclk - 100) * val_ptr->x) / 140;
	if ((temp * 140) < ((sysclk - 100) * val_ptr->x))
		temp++;
	temp += val_ptr->y;
	tb = ta = (unsigned char)((temp >> 1) & 0xff);
	if (temp & 0x1)
		ta++;

	/* (x, y) = tc_values[mode] */
	/* Tc = (ceiling((sysclk - 100) * x / 140) + y) */
	val_ptr = &tc_values[mode];
	tc = (unsigned char)(((sysclk - 100) * val_ptr->x) / 140) & 0xff;
	if ((tc * 140) < ((sysclk - 100) * val_ptr->x))
		tc++;
	tc += val_ptr->y;

	/* (x, y) = td_values[mode] */
	/* Td = ceiling(sysclk + y - 400) * x / y) */
	val_ptr = &td_values[mode];
	if ((temp = (sysclk + val_ptr->y) - 400) < 0)
		td = 0; /* It's possible that Td < 0, so we set it to 0 */
	else {
		td = (unsigned char)(((temp * val_ptr->x) / val_ptr->y) & 0xff);
		if ((td * val_ptr->y) < (temp * val_ptr->x))
			td++;
	}
	return((td << 24) | (tc << 16) | (tb << 8) | ta);
}

#endif

void __exit em86xx_isaide_exit(void)
{
#ifdef EM86XX_ISAIDE_DMA_BOUNCE_BUF
#ifdef EM86XX_BOUNCE_BUFFER_NOCACHE
        g_bounce_buf = (unsigned char *)em86xx_to_caddr((unsigned long)g_bounce_buf);
#endif
        free_pages((unsigned long)g_bounce_buf, EM86XX_ISAIDE_DMA_BOUNCE_BUF_ORDER);
#endif
        return;
}

#ifdef MODULE
module_init(em86xx_isaide_init);
module_exit(em86xx_isaide_exit);
#endif


