/*
 * $Id: physmap.c,v 1.21 2002/09/05 05:12:54 acurtis Exp $
 *
 * Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#ifdef CONFIG_TANGO2
#include <asm/tango2/emhwlib_registers_tango2.h>
#define MAX_FLASH                   4
#define XENV_MAX_FLASH_PARTITIONS   16
extern unsigned int xenv_cs_flash_size[MAX_FLASH];
extern unsigned int xenv_cs_flash_parts[MAX_FLASH];
extern unsigned int xenv_flash_parts_size[MAX_FLASH][XENV_MAX_FLASH_PARTITIONS];
extern unsigned int xenv_flash_parts_offset[MAX_FLASH][XENV_MAX_FLASH_PARTITIONS];
#endif

#ifdef CONFIG_TANGO2_SIG_BLOCK
#include <asm/tango2/sigblock.h>
#include <asm/tango2/zboot.h>
#define WINDOW_ADDR	0
#define WINDOW_SIZE	0
#define BUSWIDTH       1
#elif defined(CONFIG_TANGO2_XENV) /* Do nothing here */
#else
#define WINDOW_ADDR CONFIG_MTD_PHYSMAP_START
#define WINDOW_SIZE CONFIG_MTD_PHYSMAP_LEN
#endif

#if !defined(CONFIG_TANGO2_SIG_BLOCK) && !defined(CONFIG_TANGO2_XENV)
#define BUSWIDTH CONFIG_MTD_PHYSMAP_BUSWIDTH
#endif

#if defined(CONFIG_TANGO2_XENV)
static struct mtd_info *mymtd[MAX_FLASH] = { NULL, NULL, NULL, NULL };
#else
static unsigned long window_addr = WINDOW_ADDR;
static unsigned long window_size = WINDOW_SIZE;
static struct mtd_info *mymtd;
#endif /* CONFIG_TANGO2_XENV */

__u8 physmap_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 physmap_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 physmap_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void physmap_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

#if defined(CONFIG_TANGO2_XENV)
struct map_info physmap_map[MAX_FLASH] = {
	{
            name: "CS0: Physically mapped flash",
	    size: 0, /* Initially 0, but will be initialized later */
	    buswidth: 1, /* The default width, may be probed to change later */
	    read8: physmap_read8,
	    read16: physmap_read16,
	    read32: physmap_read32,
	    copy_from: physmap_copy_from,
	    write8: physmap_write8,
	    write16: physmap_write16,
	    write32: physmap_write32,
	    copy_to: physmap_copy_to
	},
	{
            name: "CS1: Physically mapped flash",
	    size: 0, /* Initially 0, but will be initialized later */
	    buswidth: 1, /* The default width, may be probed to change later */
	    read8: physmap_read8,
	    read16: physmap_read16,
	    read32: physmap_read32,
	    copy_from: physmap_copy_from,
	    write8: physmap_write8,
	    write16: physmap_write16,
	    write32: physmap_write32,
	    copy_to: physmap_copy_to
	},
	{
            name: "CS2: Physically mapped flash",
	    size: 0, /* Initially 0, but will be initialized later */
	    buswidth: 1, /* The default width, may be probed to change later */
	    read8: physmap_read8,
	    read16: physmap_read16,
	    read32: physmap_read32,
	    copy_from: physmap_copy_from,
	    write8: physmap_write8,
	    write16: physmap_write16,
	    write32: physmap_write32,
	    copy_to: physmap_copy_to
	},
	{
            name: "CS3: Physically mapped flash",
	    size: 0, /* Initially 0, but will be initialized later */
	    buswidth: 1, /* The default width, may be probed to change later */
	    read8: physmap_read8,
	    read16: physmap_read16,
	    read32: physmap_read32,
	    copy_from: physmap_copy_from,
	    write8: physmap_write8,
	    write16: physmap_write16,
	    write32: physmap_write32,
	    copy_to: physmap_copy_to
	},
};
#else
struct map_info physmap_map = {
	name: "Physically mapped flash",
	size: 0, /* Initially WINDOW_SIZE, but will be initialized later */
	buswidth: BUSWIDTH, /* The default width, may be probed to change later */
	read8: physmap_read8,
	read16: physmap_read16,
	read32: physmap_read32,
	copy_from: physmap_copy_from,
	write8: physmap_write8,
	write16: physmap_write16,
	write32: physmap_write32,
	copy_to: physmap_copy_to
};
#endif /* CONFIG_TANGO2_XENV */

#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
static struct mtd_partition *mtd_parts = 0;
static int                   mtd_parts_nb = 0;
#else
#if defined(CONFIG_TANGO2_XENV)
static struct mtd_partition *physmap_partitions[MAX_FLASH] = { NULL, NULL, NULL, NULL };
#else
static struct mtd_partition physmap_partitions[] = {
/* Put your own partition definitions here */
#if defined(CONFIG_TANGO2_SIG_BLOCK)
	{	name: "Partition1", size: 0, offset: 0, mask_flags: 0 },
	{	name: "Partition2", size: 0, offset: 0, mask_flags: 0 },
	{	name: "Partition3", size: 0, offset: 0, mask_flags: 0 },
	{	name: "Partition4", size: 0, offset: 0, mask_flags: 0 },
	{	name: "Partition5", size: 0, offset: 0, mask_flags: 0 }
#else
	{
		name:		"zBoot",
		size:		0x10000,
		offset:		0,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"BootStage1",
		size:		0x30000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"Linux",
		size:		0x3b0000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"User FS",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND,
	}
#endif /* CONFIG_TANGO2_SIG_BLOCK */
};
#endif 

#if !defined(CONFIG_TANGO2_XENV)
static int num_partitions = (sizeof(physmap_partitions)/sizeof(struct mtd_partition));
#endif

#endif /* CONFIG_MTD_CMDLINE_PARTS */
#endif /* CONFIG_MTD_PARTITIONS */

#if defined(CONFIG_TANGO2_XENV)
int __init init_physmap(void)
{
	static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", "map_rom", 0 };
	const char **type;
	int i, j;
	struct mtd_partition *part_ptr;
	unsigned long w_addr;
	unsigned long csconfig = gbus_read_uint32(pGBus, REG_BASE_host_interface + PB_CS_config) & 0xf;

	for (i = 0, w_addr = 0x40000000; i < MAX_FLASH; i++, w_addr += 0x04000000) {
		if ((physmap_map[i].size = xenv_cs_flash_size[i]) == 0)
			continue;
		physmap_map[i].map_priv_1 = (unsigned long)ioremap(w_addr, physmap_map[i].size);
		if (physmap_map[i].map_priv_1 == 0) {
			printk("Failed to ioremap at 0x%08lx, size 0x%08lx\n", w_addr, physmap_map[i].size);
			return -EIO;
		}
		printk(KERN_NOTICE "physmap flash device CS%d: %x at %x (%x)\n", i, 
				physmap_map[i].size, w_addr, physmap_map[i].map_priv_1);
		physmap_map[i].buswidth = ((csconfig >> i) & 0x1) ? 1 : 2;
		printk(KERN_NOTICE " detected buswidth: %d\n", physmap_map[i].buswidth);
		if ((physmap_map[i].size > 0) && (xenv_cs_flash_parts[i] > 0)) {
			physmap_partitions[i] = (struct mtd_partition *)kmalloc(sizeof(struct mtd_partition) *
					xenv_cs_flash_parts[i], GFP_KERNEL);
			if ((part_ptr = physmap_partitions[i]) != NULL) {
				for (j = 0; j < xenv_cs_flash_parts[i]; j++, part_ptr++) {
					memset(part_ptr, 0, sizeof(struct mtd_partition));
					part_ptr->size = xenv_flash_parts_size[i][j] & 0x7fffffff;
					if (xenv_flash_parts_size[i][j] & 0x80000000)
						part_ptr->mask_flags = MTD_WRITEABLE;
					part_ptr->offset = xenv_flash_parts_offset[i][j];
					part_ptr->name = (char *)kmalloc(16, GFP_KERNEL);
					if (part_ptr->name != NULL)
						sprintf(part_ptr->name, "CS%d-Part%d", i, j + 1);
				} 
			} else {
				printk("Cannot allocate memory, size 0x%x\n", 
						sizeof(struct mtd_partition) * xenv_cs_flash_parts[i]);
				return(-ENOMEM);
			}
		}

		for (type = rom_probe_types; !mymtd[i] && *type; type++) 
			mymtd[i] = do_map_probe(*type, &physmap_map[i]);

		if ((mymtd[i]) && (mymtd[i]->size != physmap_map[i].size)) {
			printk(KERN_NOTICE "Flash size mismatched (specified: 0x%x, detected: 0x%x).\n", 
					physmap_map[i].size, mymtd[i]->size);

			/* Redo ioremap */
			iounmap((void *)physmap_map[i].map_priv_1);
			physmap_map[i].size = mymtd[i]->size;
			physmap_map[i].map_priv_1 = (unsigned long)ioremap(w_addr, physmap_map[i].size);
			if (physmap_map[i].map_priv_1 == 0) {
				printk("Failed to ioremap at 0x%08lx, size 0x%08lx\n", w_addr, physmap_map[i].size);
				return -EIO;
			}
			printk(KERN_NOTICE "physmap flash device CS%d: %x at %x (remapped %x)\n", i, 
					physmap_map[i].size, w_addr, physmap_map[i].map_priv_1);
		}

		if (mymtd[i]) {
			mymtd[i]->module = THIS_MODULE;
			add_mtd_device(mymtd[i]);
			printk(KERN_NOTICE "Using physmap partition definition\n");
			add_mtd_partitions (mymtd[i], physmap_partitions[i], xenv_cs_flash_parts[i]);
		}
	}
	return(0);
}

static void __exit cleanup_physmap(void)
{
	int i, j;
	struct mtd_partition *part_ptr;
	for (i = 0; i < MAX_FLASH; i++) {
		if (mymtd[i]) {
			del_mtd_device(mymtd[i]);
			map_destroy(mymtd[i]);
		}

		if (physmap_map[i].map_priv_1) {
			iounmap((void *)physmap_map[i].map_priv_1);
			physmap_map[i].map_priv_1 = 0;
			if (physmap_partitions[i] != NULL) {
				if ((part_ptr = physmap_partitions[i]) != NULL) {
					for (j = 0; j < xenv_cs_flash_parts[i]; j++, part_ptr++) {
						if (part_ptr->name != NULL) {
							kfree(part_ptr->name);
							part_ptr->name = NULL;
						}
					}
				}
				kfree(physmap_partitions[i]);
				physmap_partitions[i] = NULL;
			}
		}
	}
}

#else /* !CONFIG_TANGO2_XENV */

int __init init_physmap(void)
{
	static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", "map_rom", 0 };
	const char **type;

#ifdef CONFIG_TANGO2_SIG_BLOCK
	extern struct signature_block *sigptr;
	if (pflash_info(sigptr, &window_addr, &window_size) < 0) {
		printk(KERN_NOTICE "invalid flash description in signature block.\n");
		return -EIO;
	}
#endif

	physmap_map.map_priv_1 = (unsigned long)ioremap(window_addr, window_size);
       	printk(KERN_NOTICE "physmap flash device: %x at %x (%x)\n", window_size, window_addr, physmap_map.map_priv_1);

	physmap_map.size = window_size;

#ifdef CONFIG_TANGO2_SIG_BLOCK
	/* If window size larger than 0, and window address is the same as CSx (x=0..3),
	 * we can then probe for the buswidth from PB_CS_config */
	if (window_size > 0) {
		unsigned long cs = 0xf;
		unsigned long csconfig = gbus_read_uint32(pGBus, REG_BASE_host_interface + PB_CS_config) & 0xf;
		switch(window_addr) {
			case 0x40000000: cs = 0;
					 break;
			case 0x44000000: cs = 1;
					 break;
			case 0x48000000: cs = 2;
					 break;
			case 0x4c000000: cs = 3;
					 break;
		}
		if (cs != 0xf) {
			/* Tango2 only supports 8 or 16 bits */
			physmap_map.buswidth = ((csconfig >> cs) & 0x1) ? 1 : 2;
			printk(KERN_NOTICE " detected buswidth: %d\n", physmap_map.buswidth);
		} else {
			/* Use whatever has been setup */
			printk(KERN_NOTICE " detected buswidth: failed, use default %d\n", physmap_map.buswidth);
		}
	} else {
		/* Use whatever has been setup */
		printk(KERN_NOTICE " detected buswidth: failed, use default %d\n", physmap_map.buswidth);
	}
#endif

	if (!physmap_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	
	mymtd = 0;
	type = rom_probe_types;
	for(; !mymtd && *type; type++) {
		mymtd = do_map_probe(*type, &physmap_map);
	}
	if (mymtd) {
		mymtd->module = THIS_MODULE;

		add_mtd_device(mymtd);

#if defined(CONFIG_TANGO2_SIG_BLOCK)
		{
			int i, idx;
			unsigned long size, offset;
			for (idx = 0, i = 1; i <= pflash_num_partitions(); i++) {
				if (pflash_partition_info(sigptr, i, &offset, &size) < 0)
					continue;
				else if ((size & 0x7fffffff) == 0)      /* Skip the empty partition */
					continue;
				physmap_partitions[idx].size = size & 0x7fffffff;
				physmap_partitions[idx].offset = offset;
				physmap_partitions[idx].mask_flags = ((size & 0x80000000UL) ? MTD_WRITEABLE : 0);
				idx++;
			}
			num_partitions = idx;
		}
#endif

#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
		mtd_parts_nb = parse_cmdline_partitions(mymtd, &mtd_parts, 
							"phys");
		if (mtd_parts_nb > 0)
		{
			printk(KERN_NOTICE 
			       "Using command line partition definition\n");
			add_mtd_partitions (mymtd, mtd_parts, mtd_parts_nb);
		}
#else
		if (num_partitions != 0) 
		{
			printk(KERN_NOTICE 
			       "Using physmap partition definition\n");
			add_mtd_partitions (mymtd, physmap_partitions, num_partitions);
		}

#endif
#endif
		return 0;
	}

	iounmap((void *)physmap_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_physmap(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (physmap_map.map_priv_1) {
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
	}
}
#endif /* CONFIG_TANGO2_XENV */

module_init(init_physmap);
module_exit(cleanup_physmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
