#include "kernelheaders.h"
#include "common.h"

#if (EM86XX_CHIP==EM86XX_CHIPID_TANGO2)
#include <asm/tango2/hardware.h>
#endif

struct gbus_fifo *gbus_fifo_open(struct gbus *pgbus, RMuint32 data_address, RMuint32 data_size, RMuint32 fifo_address)
{
	struct gbus_fifo *fifo = (struct gbus_fifo *) fifo_address;
	
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->base), data_address);
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->size), data_size);
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->rd), 0);
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->wr), 0);
	
	return fifo;
}

void gbus_fifo_get_pointer(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *base, RMuint32 *size, RMuint32 *rd_ptr, RMuint32 *wr_ptr)
{
	*base = gbus_read_uint32(pgbus, (RMuint32) &(fifo->base));
	*size = gbus_read_uint32(pgbus, (RMuint32) &(fifo->size));
	*rd_ptr = gbus_read_uint32(pgbus, (RMuint32) &(fifo->rd));
	*wr_ptr = gbus_read_uint32(pgbus, (RMuint32) &(fifo->wr));
}

RMuint32 gbus_fifo_get_readable_size(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *rd_ptr1, RMuint32 *rd_size1, RMuint32 *rd_ptr2)
{
	RMuint32 base, size, rd, wr;
	
	gbus_fifo_get_pointer(pgbus, fifo, &base, &size, &rd, &wr);
	
	*rd_ptr1 = base + rd;
	
	if (wr >= rd) {
		*rd_size1 = wr - rd;
		*rd_ptr2 = 0;
		return (*rd_size1);
	}
	else {
		*rd_size1 = size - rd;
		*rd_ptr2 = base;
		return (*rd_size1 + wr);
	}
}

RMuint32 gbus_fifo_get_writable_size(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *wr_ptr1, RMuint32 *wr_size1, RMuint32 *wr_ptr2)
{
	RMuint32 base, size, rd, wr;
	
	gbus_fifo_get_pointer(pgbus, fifo, &base, &size, &rd, &wr);
	
	*wr_ptr1 = base + wr;
	
	if (wr >= rd) {
		if (rd > 0) {
			*wr_size1 = size - wr;
			*wr_ptr2 = base;
			return (*wr_size1 + rd - 1);
		}
		else {
			*wr_size1 = size - 1 - wr;
			*wr_ptr2 = 0;
			return (*wr_size1);
		}			
	}
	else {
		*wr_size1 = rd - 1 - wr;
		*wr_ptr2 = 0;
		return (*wr_size1);
	}
}

RMuint32 gbus_fifo_incr_write_ptr(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 incr)
{
	RMuint32 base, size, rd, wr;
	
	gbus_fifo_get_pointer(pgbus, fifo, &base, &size, &rd, &wr);
	
	wr += incr;
	if (wr >= size)
		wr -= size;
	
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->wr), wr);
		
	return wr + base;
}

RMuint32 gbus_fifo_incr_read_ptr(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 incr)
{
	RMuint32 base, size, rd, wr;
	
	gbus_fifo_get_pointer(pgbus, fifo, &base, &size, &rd, &wr);
	
	rd += incr;
	if (rd >= size)
		rd -= size;
		
	gbus_write_uint32(pgbus, (RMuint32) &(fifo->rd), rd);
	
	return rd + base;
}

static unsigned long *tasklet_status = (unsigned long *) NULL;
static struct tasklet_struct *tasklet = (struct tasklet_struct *) NULL;
static unsigned long tasklet_mask = 0;

static void tophalf(int irq, void *dev_id, struct pt_regs *regs) 
{
	unsigned long status;
	unsigned long mask;
	unsigned long flags;

	status = gbus_read_uint32(pGBus, REG_BASE_cpu_block + CPU_irq_softset);

	if ((status & SOFT_IRQ_ORIGIN_PT110) == 0) {
		printk("Invalid software interrupt 0x%08lx\n", status);
		return;
	}
	
 	gbus_mutex_lock(pGBus, SOFT_IRQ_MUTEX_IRQ);

	status = gbus_read_uint32(pGBus, HOST_INTERRUPT_STATUS);
	mask = status & (tasklet_mask);
	status &= ~(tasklet_mask); 
	gbus_write_uint32(pGBus, HOST_INTERRUPT_STATUS, status);
	if (status == 0)
		gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_irq_softclr, SOFT_IRQ_ORIGIN_PT110);
	
	gbus_mutex_unlock(pGBus, SOFT_IRQ_MUTEX_IRQ);

	if ((mask) && (tasklet)) {
		*tasklet_status |= mask;
		tasklet_schedule(tasklet);
	}
}

void mumk_register_tasklet(struct gbus *pgbus, struct kc_tasklet_struct *kctasklet, RMuint32 *irq_status, RMuint32 mask)
{
	request_irq(IRQ_SOFTINT, tophalf, SA_INTERRUPT | SA_SHIRQ, "gnet0", &(Gtable[0]));
	tasklet_status = irq_status;
	tasklet = (struct tasklet_struct *)kctasklet;
	tasklet_mask = mask;
	printk("mumk_register_tasklet: done\n");
}

void mumk_unregister_tasklet(struct gbus *pgbus, struct kc_tasklet_struct *kctasklet)
{
	tasklet_mask = 0;
	tasklet = (struct tasklet_struct *) NULL; 
	tasklet_status = (unsigned long *) NULL; 
	free_irq(IRQ_SOFTINT, &(Gtable[0]));
	printk("mumk_unregister_tasklet: cannot unregister\n");
}

