#ifndef __EMHWLIB_H__
#define __EMHWLIB_H__

#if (defined CONFIG_ARCH_MAMBO)
#define EM86XX_CHIP EM86XX_CHIPID_MAMBOLIGHT
#elif (defined CONFIG_ARCH_TANGO)
#define EM86XX_CHIP EM86XX_CHIPID_TANGOLIGHT
#elif (defined CONFIG_TANGO2)
#define EM86XX_CHIP EM86XX_CHIPID_TANGO2
#endif

#define TRUE 1
#define FALSE 0

#define RMascii char 
#define RMbool RMuint8
#define RMint32 long
#define RMuint32 unsigned long
#define RMuint8 unsigned char

#define ENABLE TRUE, __FILE__, __LINE__
#define DISABLE FALSE, __FILE__, __LINE__

void RMDBGLOG_implementation(RMbool active,const RMascii *filename,RMint32 line,const RMascii *text,...);
void RMDBGPRINT_implementation(RMbool active,const RMascii *filename,RMint32 line,const RMascii *text,...);

#define RMASSERT(x)
#define RMDBGLOG(x) RMDBGLOG_implementation x
#define RMDBGPRINT(x) RMDBGPRINT_implementation x
#define gbus_logger(x,y,z)

#define RMmin(x,y) (((x) < (y)) ? (x) : (y))

#define RMmustBeEqual(x,y,seed)                 \
typedef char XXX ## seed ## LeftIsBiggerNow[(y)-(x)];           \
typedef char XXX ## seed ## LeftIsSmallerNow[(x)-(y)];
#define RMleftMustBeSmaller(x,y,seed)                 \
typedef RMascii XXX ## seed ## LeftIsBiggerNow[(y)-(x)];           

#define SOFT_IRQ_ORIGIN_PT110      (1<<4)

#define LOG2_LLAD_GNET0RX                       5
#define LOG2_LLAD_GNET1RX                       6
#define LLAD_GNET0RX                       (1 << LOG2_LLAD_GNET0RX)
#define LLAD_GNET1RX                       (1 << LOG2_LLAD_GNET1RX)

struct gbus;
struct llad;

#if (EM86XX_CHIP!=EM86XX_CHIPID_TANGO2)

#include "emhwlib_dram.h"
#include "emhwlib_resources.h"
#include "emhwlib_registers.h"
#define EM86XX_MODE EM86XX_MODEID_STANDALONE

#define pGBus ((struct gbus *) 1)
#define pLLAD ((struct llad *) 1)

static inline struct llad *llad_open(char *dev)
{
	if (*dev == '0')
		return pLLAD;
	else
		return (struct llad *) NULL;
}

static inline void llad_close(struct llad *pllad)
{
}

static inline struct gbus *gbus_open(struct llad *pllad)
{
	return pGBus;
}

static inline void gbus_close(struct gbus *pgbus)
{
}

static inline RMuint8  gbus_read_uint8 (struct gbus *h, RMuint32 byte_address)
{
	return *((volatile RMuint8 *) byte_address);
}

static inline RMuint32 gbus_read_uint32(struct gbus *h, RMuint32 byte_address)
{
	return *((volatile RMuint32 *) byte_address);
}

static inline void gbus_write_uint8 (struct gbus *h, RMuint32 byte_address, RMuint8  data)
{
	*((volatile RMuint8 *) byte_address) = data;
}

static inline void gbus_write_uint32(struct gbus *h, RMuint32 byte_address, RMuint32 data)
{
	*((volatile RMuint32 *) byte_address) = data;
}

#else /* tango2 */


#include "asm/tango2/emhwlib_dram.h"
#include "asm/tango2/emhwlib_lram.h"
#include "asm/tango2/emhwlib_registers_tango2.h"
#include "asm/tango2/emhwlib_resources_tango2.h"
#include "asm/tango2/emhwlib_resources_shared.h"

#define pLLAD ((struct llad *) 1)

static inline struct llad *llad_open(char *dev)
{
	if (*dev == '0')
		return pLLAD;
	else
		return (struct llad *) NULL;
}

static inline void llad_close(struct llad *pllad)
{
}

static inline struct gbus *gbus_open(struct llad *pllad)
{
	return pGBus;
}

static inline void gbus_close(struct gbus *pgbus)
{
}

#ifndef __KERNEL__
static inline RMuint8  gbus_read_uint8 (struct gbus *h, RMuint32 byte_address)
{
	return *((volatile RMuint8 *) KSEG1ADDR(byte_address));
}

static inline void gbus_write_uint8 (struct gbus *h, RMuint32 byte_address, RMuint8  data)
{
	*((volatile RMuint8 *) KSEG1ADDR(byte_address)) = data;
}
#endif
#endif

static inline void gbus_read_data8 (struct gbus *h, RMuint32 byte_address, RMuint8  *data, RMuint32 count)
{
	RMuint32 i;

	for (i=0 ; i<count ; i++) {
		*(data+i) = gbus_read_uint8(h, byte_address + i);
	}
}

static inline void gbus_write_data8 (struct gbus *h, RMuint32 byte_address, RMuint8  *data, RMuint32 count)
{
	RMuint32 i;

	for (i=0 ; i<count ; i++) {
		gbus_write_uint8(h, byte_address + i, *(data+i));
	}
}

#define gbus_save_flags_clf(x)               \
        {                               \
                RMuint32 tmp;           \
                __asm__(                \
                "mrs    %0, cpsr\n"     \
                "orr    %1, %0, #64\n"  \
                "msr    cpsr_c, %1\n"   \
                : "=r" (x), "=r" (tmp)  \
                :                       \
                : "memory");            \
        }

/* 
 * Restore the saved value back to CPSR register.
 */
#define gbus_restore_flags(x)                \
        __asm__(                        \
        "msr    cpsr_c, %0\n"           \
        :                               \
        : "r" (x))

struct gbus_mutex;
static inline void gbus_mutex_lock(struct gbus *pgbus, struct gbus_mutex *mutex)
{
	while (gbus_read_uint32(pgbus, (RMuint32) mutex));
}

static inline void gbus_mutex_unlock(struct gbus *pgbus, struct gbus_mutex *mutex)
{
	gbus_write_uint32(pgbus, (RMuint32) mutex, 0);
}


struct gbus_fifo {
	RMuint32 base;
	RMuint32 size;
	RMuint32 rd;
	RMuint32 wr;
};

struct gbus_fifo *gbus_fifo_open(struct gbus *pgbus, RMuint32 data_address, RMuint32 data_size, RMuint32 fifo_address);
void gbus_fifo_get_pointer(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *base, RMuint32 *size, RMuint32 *rd_ptr, RMuint32 *wr_ptr);
RMuint32 gbus_fifo_get_readable_size(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *rd_ptr1, RMuint32 *rd_size1, RMuint32 *rd_ptr2);
RMuint32 gbus_fifo_get_writable_size(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 *wr_ptr1, RMuint32 *wr_size1, RMuint32 *wr_ptr2);
RMuint32 gbus_fifo_incr_write_ptr(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 incr);
RMuint32 gbus_fifo_incr_read_ptr(struct gbus *pgbus, struct gbus_fifo *fifo, RMuint32 incr);

struct kc_tasklet_struct;
void mumk_register_tasklet(struct gbus *pgbus, struct kc_tasklet_struct *tasklet, RMuint32 *irq_status, RMuint32 mask);
void mumk_unregister_tasklet(struct gbus *pgbus, struct kc_tasklet_struct *kctasklet);


#endif // __EMHWLIB_H__
