/*****************************************
 Copyright  2001-2003	
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/

/*
  @file   net.c
  See licensing details in LICENSING file  
*/

#include "kernelheaders.h"
#include "common.h"

// to enable or disable the debug messages of this source file, put 1 or 0 below
#if 0
#define LOCALDBG ENABLE
#else
#define LOCALDBG DISABLE
#endif

void gnet_tasklet(unsigned long cookie)
{
	struct gnetprivate *pG=(struct gnetprivate *)cookie;
	RMuint32 rxcount=0;
	
 	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_PD10:GBUS_LOGGER_PD01,0);
 	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_RX0:GBUS_LOGGER_RX1,1);

	while (1) {
		RMuint32 ptr1;
		RMuint32 size1;
		RMuint32 ptr2;
		RMuint32 size;
		RMuint32 blocklen;
		
		RMuint32 len;
		struct sk_buff *skb;
		RMuint8 *dst;
		
		size=gbus_fifo_get_readable_size(pG->pgbus,pG->rxfifo,&ptr1,&size1,&ptr2);

		/* tries to read gnet header */
		if (size<4) 
			break;
		else {
			len=gbus_read_uint32(pG->pgbus,ptr1);
		}
		
		/* blocklen is always a multiple of 4 bytes */
		blocklen=(4+len+3) & ~3;
		
		if (size<blocklen) {
			PG_RMDBGLOG((ENABLE,"gnet_tasklet: algorithmic fifo error (rxcount=%d size=%d len=%d blocklen=%d)\n",
				     rxcount,size,len,blocklen));
			break;
		}
		
		ptr1+=4;
		size1-=4;

		/* allocates 2 bytes more (skipped bytes) */
		skb=dev_alloc_skb(len+2);
		if (skb == NULL) {
			pG->netstats.rx_dropped++;
			
			PG_RMDBGLOG((LOCALDBG,"gnet_tasklet: memory squeeze (dropped %d)\n",pG->netstats.rx_dropped));
			break;
		}

		/* skip 2 bytes so that IP is aligned on 16 Bytes boundary */
		skb_reserve(skb,2);

		/* prepare skb to receive len bytes */
		dst = skb_put(skb,len);
		
		while (len > 0) {
			RMuint32 xfer_size;
			
			if (size1 == 0) {
				size1 = len;
				ptr1 = ptr2;
			}
			
			xfer_size = RMmin(len, size1);
			gbus_read_data8(pG->pgbus, ptr1, dst, xfer_size);
			
			dst += xfer_size;
			len -= xfer_size;
			size1 -= xfer_size;
		}

		gbus_fifo_incr_read_ptr(pG->pgbus,pG->rxfifo,blocklen);
		
		skb->dev = &pG->netdev;
		skb->protocol = eth_type_trans(skb,&pG->netdev);
		skb->ip_summed = CHECKSUM_UNNECESSARY;
		pG->netdev.last_rx = jiffies;
		pG->netstats.rx_bytes += skb->len;
		pG->netstats.rx_packets++;
		netif_rx(skb);
		
		PG_RMDBGLOG((LOCALDBG,"gnet_tasklet: rx skb 0x%08lx\n",skb));
		
		rxcount++;
		
		{
			RMuint32 pending;

			pending=gbus_read_uint32(pG->pgbus,(which==0)?RXC0:RXC1);
			gbus_write_uint32(pG->pgbus,(which==0)?RXC0:RXC1,pending+1);
		}
	}
	
	PG_RMDBGLOG((DISABLE,"gnet_tasklet: %d skb's processed\n",rxcount));

	gbus_logger(pG->pgbus,GBUS_LOGGER_COUNTER0,gbus_read_uint32(pG->pgbus,TXC0)-gbus_read_uint32(pG->pgbus,RXC1));
	gbus_logger(pG->pgbus,GBUS_LOGGER_COUNTER1,gbus_read_uint32(pG->pgbus,TXC1)-gbus_read_uint32(pG->pgbus,RXC0));

	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_RX0:GBUS_LOGGER_RX1,0);
	return;
}

static int gnet_header(struct sk_buff *skb, struct net_device *dev,
                unsigned short type, void *daddr, void *saddr,
		       unsigned int len)
{
	struct ethhdr *eth = (struct ethhdr *)skb_push(skb,ETH_HLEN);
	struct gnetprivate *pG=(struct gnetprivate *)dev->priv;
	
	eth->h_proto = htons(type);
	memcpy(eth->h_source, saddr ? saddr : dev->dev_addr, dev->addr_len);
	memcpy(eth->h_dest,   daddr ? daddr : dev->dev_addr, dev->addr_len);
	eth->h_dest[ETH_ALEN-1]   = 1-(pG-Gtable);
	PG_RMDBGLOG((DISABLE,"gnet_header: s=%02x:%02x:%02x:%02x:%02x:%02x d=%02x:%02x:%02x:%02x:%02x:%02x\n",
		     eth->h_source[0],
		     eth->h_source[1],
		     eth->h_source[2],
		     eth->h_source[3],
		     eth->h_source[4],
		     eth->h_source[5],
		     eth->h_dest[0],
		     eth->h_dest[1],
		     eth->h_dest[2],
		     eth->h_dest[3],
		     eth->h_dest[4],
		     eth->h_dest[5]));

	return (dev->hard_header_len);
}

static int gnet_tx_packet(struct sk_buff *skb, struct net_device *dev)
{
	struct gnetprivate *pG=(struct gnetprivate *)dev->priv;
	int rc;
	RMuint32 flags;
	RMuint32 ptr1;
	RMuint32 size1;
	RMuint32 ptr2;
	RMuint32 size;

	RMuint32 blocklen, len;
	RMuint8 *src;
	
	save_flags(flags); cli();
	
	gbus_logger(pG->pgbus,GBUS_LOGGER_COUNTER0,gbus_read_uint32(pG->pgbus,TXC0)-gbus_read_uint32(pG->pgbus,RXC1));
	gbus_logger(pG->pgbus,GBUS_LOGGER_COUNTER1,gbus_read_uint32(pG->pgbus,TXC1)-gbus_read_uint32(pG->pgbus,RXC0));
	
	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_TX0:GBUS_LOGGER_TX1,1);
	
	if (skb==NULL) {
		PG_RMDBGLOG((LOCALDBG,"gnet_tx_packet: skb==NULL!\n"));
		rc=0;
		goto out;
	}
	
	/* blocklen is (data length + gnet header length) + padding size to align on 4 bytes boundary */
	blocklen = (skb->len + 4 + 3) & ~3;

	size=gbus_fifo_get_writable_size(pG->pgbus,pG->txfifo,&ptr1,&size1,&ptr2);
	
	if (size<blocklen) {
		PG_RMDBGLOG((LOCALDBG,"gnet_tx_packet: not enough room, rejecting 0x%08lx\n",skb));
		rc=1;
		goto out;
	}

	dev->trans_start = jiffies;
		
	if (which==2) {
		// kind of trick when two interfaces sit on the same host ;-)
		
		struct iphdr *ih;
		u32 *saddr, *daddr;
		
		// perform some changes
		ih = (struct iphdr *)(skb->data+sizeof(struct ethhdr));
		saddr = &ih->saddr;
		daddr = &ih->daddr;
		PG_RMDBGLOG((LOCALDBG,"gnet_tx_packet: from %d.%d.%d.%d to %d.%d.%d.%d\n",
			     (int)((u8 *)saddr)[0],
			     (int)((u8 *)saddr)[1],
			     (int)((u8 *)saddr)[2],
			     (int)((u8 *)saddr)[3],
			     (int)((u8 *)daddr)[0],
			     (int)((u8 *)daddr)[1],
			     (int)((u8 *)daddr)[2],
			     (int)((u8 *)daddr)[3]));
		((u8 *)saddr)[2] = 1-((u8 *)saddr)[2];
		((u8 *)daddr)[2] = 1-((u8 *)daddr)[2];
		
		ih->check = 0;         /* and rebuild the checksum (ip needs it) */
		ih->check = ip_fast_csum((unsigned char *)ih,ih->ihl);
	}

	len = skb->len;
	src = skb->data;
	
	/* writes gnet header */
	gbus_write_uint32(pG->pgbus, ptr1, len);
	ptr1 += 4;
	size1 -= 4;
	
	while (len > 0) {
		RMuint32 xfer_size;
		
		if (size1 == 0) {
			size1 = len;
			ptr1 = ptr2;
		}
		
		xfer_size = RMmin(len, size1);
		gbus_write_data8(pG->pgbus, ptr1, src, xfer_size);
		
		src += xfer_size;
		len -= xfer_size;
		size1 -= xfer_size;
	}
	
	gbus_fifo_incr_write_ptr(pG->pgbus,pG->txfifo,blocklen);

	pG->netstats.tx_bytes += skb->len;
	pG->netstats.tx_packets++;
	dev_kfree_skb(skb);
	
	PG_RMDBGLOG((LOCALDBG,"gnet_tx_packet: tx skb 0x%08lx\n",skb));
	
	{
		RMuint32 pending;
		
		pending=gbus_read_uint32(pG->pgbus,(which==0)?TXC0:TXC1);
		gbus_write_uint32(pG->pgbus,(which==0)?TXC0:TXC1,pending+1);
	}
	
	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_PD01:GBUS_LOGGER_PD10,1);
	
	// raise interrupt
	switch (which) {
	case 0: // uclinux
		RMASSERT(pG==Gtable);
#if ((EM86XX_CHIP==EM86XX_CHIPID_MAMBO) || (EM86XX_CHIP==EM86XX_CHIPID_MAMBOLIGHT))
		gbus_write_uint32(pG->pgbus, REG_BASE_host_interface+PCI_IRQ,
				  LLAD_GNET1RX);
#elif (EM86XX_CHIP==EM86XX_CHIPID_TANGOLIGHT) || (EM86XX_CHIP==EM86XX_CHIPID_TANGO15) || (EM86XX_CHIP==EM86XX_CHIPID_TANGO2)
		gbus_write_uint32(pG->pgbus, REG_BASE_host_interface+PCI_irq_set,
				  LLAD_GNET1RX);
#else
#error EM86XX_CHIP is not set in RMCFLAGS: refer to emhwlib/include/emhwlib_chips.h.
#endif
		break;
		
	case 1: // host
		RMASSERT(pG==Gtable+1);
		gbus_mutex_lock(pG->pgbus, SOFT_IRQ_MUTEX_IRQ);
		gbus_mutex_lock(pG->pgbus, SOFT_IRQ_MUTEX_FIQ);
		gbus_write_uint32(pG->pgbus, HOST_INTERRUPT_STATUS,
				  LLAD_GNET0RX);
		gbus_write_uint32(pG->pgbus, REG_BASE_cpu_block+CPU_irq_softset,
				  SOFT_IRQ_ORIGIN_PT110);
		gbus_mutex_unlock(pG->pgbus, SOFT_IRQ_MUTEX_FIQ);
		gbus_mutex_unlock(pG->pgbus, SOFT_IRQ_MUTEX_IRQ);
		break;
		
	case 2: // both interfaces on uclinux or host
#if (EM86XX_MODE==EM86XX_MODEID_WITHHOST) 
		
		// can't use gbus_set_interrupt because mutex is taken twice
#if ((EM86XX_CHIP==EM86XX_CHIPID_MAMBO) || (EM86XX_CHIP==EM86XX_CHIPID_MAMBOLIGHT))
		gbus_write_uint32(pG->pgbus, REG_BASE_host_interface+PCI_IRQ,
				  (pG==Gtable)?LLAD_GNET1RX:LLAD_GNET0RX);
#elif (EM86XX_CHIP==EM86XX_CHIPID_TANGOLIGHT) || (EM86XX_CHIP==EM86XX_CHIPID_TANGO15) || (EM86XX_CHIP==EM86XX_CHIPID_TANGO2)
		gbus_write_uint32(pG->pgbus, REG_BASE_host_interface+PCI_irq_set,
				  (pG==Gtable)?LLAD_GNET1RX:LLAD_GNET0RX);
#else
#error EM86XX_CHIP is not set in RMCFLAGS: refer to emhwlib/include/emhwlib_chips.h.
#endif
		
#elif (EM86XX_MODE==EM86XX_MODEID_STANDALONE)
		gbus_mutex_lock(pG->pgbus, SOFT_IRQ_MUTEX_IRQ);
		gbus_mutex_lock(pG->pgbus, SOFT_IRQ_MUTEX_FIQ);
		// can't use gbus_set_interrupt because undef is not handled in kernel mode
		gbus_write_uint32(pG->pgbus, HOST_INTERRUPT_STATUS,
				  (pG==Gtable)?LLAD_GNET1RX:LLAD_GNET0RX);
		gbus_write_uint32(pG->pgbus, REG_BASE_cpu_block+CPU_irq_softset,
				  SOFT_IRQ_ORIGIN_PT110);
		gbus_mutex_unlock(pG->pgbus, SOFT_IRQ_MUTEX_FIQ);
		gbus_mutex_unlock(pG->pgbus, SOFT_IRQ_MUTEX_IRQ);
#else //EM86XX_MODE
		NOTCOMPILABLE;
#endif //EM86XX_MODE
	default:
		break;
	}
	
	gbus_logger(pG->pgbus,(which==0)?GBUS_LOGGER_TX0:GBUS_LOGGER_TX1,0);

	rc=0;

 out:
	restore_flags(flags);
	return rc;
}

static int gnet_open(struct net_device *dev)
{
	struct gnetprivate *pG=(struct gnetprivate *)dev->priv;

	MOD_INC_USE_COUNT;

	PG_RMDBGLOG((LOCALDBG,"gnet_open: begun\n"));
	
	// mac address 00:07:07:07:07:last-digit-of-uclinux-ip:index

	/*
	  ANY CHANGE OF MAC ADDRESS OF THE SAME IP INTERFACE WILL CAUSE 
	  MAJOR PROBLEMS SINCE THE PREVIOUS ONE HAS BEEN CACHED BY ALL COMPUTERS OF THE LAN
	 */
	*(dev->dev_addr)=0;
	memset(dev->dev_addr+1, 0x07, ETH_ALEN-3);
	*(dev->dev_addr+ETH_ALEN-2)=gbus_read_uint8(pG->pgbus, REG_BASE_cpu_block+LR_GNET_MAC);
	*(dev->dev_addr+ETH_ALEN-1)=pG-Gtable;

	memset(&pG->netstats,0,sizeof(struct net_device_stats));

	netif_start_queue(dev);

	PG_RMDBGLOG((LOCALDBG,"gnet_open: done\n"));

	return 0;
}

static int gnet_close(struct net_device *dev)
{
	struct gnetprivate *pG=(struct gnetprivate *)dev->priv;

	PG_RMDBGLOG((LOCALDBG,"gnet_close: begun\n"));

	netif_stop_queue (dev);

	PG_RMDBGLOG((LOCALDBG,"gnet_close: done\n"));

	MOD_DEC_USE_COUNT;

	return 0;
}

static struct net_device_stats *gnet_get_stats(struct net_device *dev)
{
	struct gnetprivate *pG=(struct gnetprivate *)dev->priv;

	return &pG->netstats;
}

int net_init(struct gnetprivate *pG)
{
	struct net_device *dev=&pG->netdev;

	PG_RMDBGLOG((LOCALDBG,"net_init: begun\n"));

	SET_MODULE_OWNER(dev);
	
	memset(dev,0,sizeof(struct net_device));

	sprintf(dev->name,"gnet%ld",pG-Gtable);
	
	ether_setup(dev);
	dev->hard_start_xmit	 = gnet_tx_packet;
	dev->open		 = gnet_open;
	dev->stop		 = gnet_close;
	dev->get_stats 		 = gnet_get_stats;
	dev->hard_header         = gnet_header;
	dev->hard_header_cache = NULL;      /* Disable caching */
	dev->flags	         |= IFF_NOARP|IFF_POINTOPOINT;
	//	dev->tx_queue_len = 100;

	dev->priv = pG;
	PG_RMDBGLOG((LOCALDBG,"net_init: done\n"));

	return 0;
}
