/*****************************************
 Copyright  2001-2003	
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/

/*
  @file   probe.c
  See licensing details in LICENSING file  
*/

#include "kernelheaders.h"
#include "common.h"

// to enable or disable the debug messages of this source file, put 1 or 0 below
#if 1
#define LOCALDBG ENABLE
#else
#define LOCALDBG DISABLE
#endif

#if defined(CONFIG_TANGO2_XENV)
extern int tango2_gnet_enabled(void); 
#endif

int number_of_probed_gnets;
struct gnetprivate Gtable[2];

MODULE_DESCRIPTION("gbus network driver");
MODULE_AUTHOR("Emmanuel Michon <emmanuel_michon@gnet.fr>");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Proprietary");  // parsed keyword, don't change!
#endif // MODULE_LICENSE

#ifdef CONFIG_GNET
int which=0;
#else
int which=-1; // 0: probe first (uclinux side). 1: probe second (host side). 2: probe both on uclinux / both on host
#endif

MODULE_PARM(which,"i");

static RMuint32 mask=0;
static struct gbus *pgbus;
static unsigned long master_status;
static struct tasklet_struct master_tasklet;

static void master_routine(unsigned long cookie)
{
	struct gnetprivate *pG=(struct gnetprivate *)cookie;

	if (master_status&LLAD_GNET0RX) 
		gnet_tasklet((unsigned long)pG);
	
	if (master_status&LLAD_GNET1RX) 
		gnet_tasklet((unsigned long)(pG+1));
}

static int __init gnet_init_module(void)
{  
	struct gbus_fifo *f0,*f1;

#if defined(CONFIG_TANGO2_XENV)
	if (tango2_gnet_enabled() == 0) 
		return 0;
#endif
	RMDBGLOG((LOCALDBG,"init_module: begun\n"));
	
	if (which==-1) return -EINVAL;

	pgbus=gbus_open(llad_open("0"));
	gbus_write_uint32(pgbus,TXC0,0);
	gbus_write_uint32(pgbus,TXC1,0);
	gbus_write_uint32(pgbus,RXC0,0);
	gbus_write_uint32(pgbus,RXC1,0);
	f0=gbus_fifo_open(pgbus,F0DATA,DATALEN,F0);
	f1=gbus_fifo_open(pgbus,F1DATA,DATALEN,F1);

	if (which!=1) {
		mask|=LLAD_GNET0RX;
		gnetk_init(&Gtable[0],pgbus,f0,f1);
		number_of_probed_gnets++;
	}
	if (which!=0) {
		mask|=LLAD_GNET1RX;
		gnetk_init(&Gtable[1],pgbus,f1,f0);
		number_of_probed_gnets++;
	}
	
	tasklet_init(&master_tasklet,master_routine,(unsigned long)Gtable);
	mumk_register_tasklet(pgbus,(struct kc_tasklet_struct *)&master_tasklet,&master_status, (LLAD_GNET0RX|LLAD_GNET1RX));
	
	RMDBGLOG((ENABLE,"init_module: found %d interface%s, done\n",
		  number_of_probed_gnets,
		  (number_of_probed_gnets>1)?"s":""));
	
	return 0;
}

static void __exit gnet_cleanup_module(void)
{
#if defined(CONFIG_TANGO2_XENV)
	if (tango2_gnet_enabled() == 0) 
		return;
#endif
	RMDBGLOG((LOCALDBG,"cleanup_module: begun\n"));

	tasklet_disable(&master_tasklet);
	mumk_unregister_tasklet(pgbus,(struct kc_tasklet_struct *)&master_tasklet);
	
	if (which!=1) gnetk_cleanup(&Gtable[0]);
	if (which!=0) gnetk_cleanup(&Gtable[1]);
	number_of_probed_gnets=0;
	
	// hack: should call gbus_fifo_close here.
	
	RMDBGLOG((ENABLE,"cleanup_module: done\n"));
}

module_init(gnet_init_module);
module_exit(gnet_cleanup_module);
